#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base class for all ZenML deployers."""

import secrets
import string
import time
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from uuid import UUID

import requests

from zenml.analytics.enums import AnalyticsEvent
from zenml.analytics.utils import track_handler
from zenml.client import Client
from zenml.config import DeploymentDefaultEndpoints
from zenml.config.base_settings import BaseSettings
from zenml.constants import (
    ENV_ZENML_ACTIVE_PROJECT_ID,
    ENV_ZENML_ACTIVE_STACK_ID,
)
from zenml.deployers.exceptions import (
    DeployerError,
    DeploymentAlreadyExistsError,
    DeploymentDeployerMismatchError,
    DeploymentDeprovisionError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
    DeploymentSnapshotMismatchError,
    DeploymentTimeoutError,
)
from zenml.enums import DeploymentStatus, StackComponentType
from zenml.logger import get_logger
from zenml.models import (
    DeploymentOperationalState,
    DeploymentRequest,
    DeploymentResponse,
    DeploymentUpdate,
    PipelineSnapshotResponse,
)
from zenml.orchestrators.utils import get_config_environment_vars
from zenml.stack import StackComponent
from zenml.stack.flavor import Flavor
from zenml.stack.stack_component import StackComponentConfig
from zenml.utils.uuid_utils import is_valid_uuid

if TYPE_CHECKING:
    from zenml.stack import Stack

logger = get_logger(__name__)

DEFAULT_DEPLOYMENT_LCM_TIMEOUT = 600


class BaseDeployerSettings(BaseSettings):
    """Base settings for all deployers."""

    auth_key: Optional[str] = None
    generate_auth_key: bool = False
    lcm_timeout: int = DEFAULT_DEPLOYMENT_LCM_TIMEOUT


class BaseDeployerConfig(StackComponentConfig):
    """Base config for all deployers."""


class BaseDeployer(StackComponent, ABC):
    """Base class for all ZenML deployers.

    The deployer serves three major purposes:

    1. It contains all the stack related configuration attributes required to
    interact with the remote pipeline deployment tool, service or platform (e.g.
    hostnames, URLs, references to credentials, other client related
    configuration parameters).

    2. It implements the life-cycle management for deployments, including
    discovery, creation, deletion and updating.

    3. It acts as a ZenML deployment registry, where every pipeline
    deployment is stored as a database entity through the ZenML Client. This
    allows the deployer to keep track of all externally running pipeline
    deployments and to manage their lifecycle.
    """

    @property
    def config(self) -> BaseDeployerConfig:
        """Returns the `BaseDeployerConfig` config.

        Returns:
            The configuration.
        """
        return cast(BaseDeployerConfig, self._config)

    @classmethod
    def get_active_deployer(cls) -> "BaseDeployer":
        """Get the deployer registered in the active stack.

        Returns:
            The deployer registered in the active stack.

        Raises:
            TypeError: if a deployer is not part of the
                active stack.
        """
        client = Client()
        deployer = client.active_stack.deployer
        if not deployer or not isinstance(deployer, cls):
            raise TypeError(
                "The active stack needs to have a deployer "
                "component registered to be able to deploy pipelines. "
                "You can create a new stack with a deployer component "
                "or update your active stack to add this component, e.g.:\n\n"
                "  `zenml deployer register ...`\n"
                "  `zenml stack register <STACK-NAME> -D ...`\n"
                "  or:\n"
                "  `zenml stack update -D ...`\n\n"
            )

        return deployer

    def _update_deployment(
        self,
        deployment: DeploymentResponse,
        operational_state: DeploymentOperationalState,
    ) -> DeploymentResponse:
        """Update an existing deployment instance with the operational state.

        Args:
            deployment: The deployment to update.
            operational_state: The operational state of the deployment.

        Returns:
            The updated deployment.
        """
        client = Client()
        if operational_state.status == DeploymentStatus.ABSENT:
            # Erase the URL and metadata for absent deployments
            operational_state.url = ""
            operational_state.metadata = {}

        return client.zen_store.update_deployment(
            deployment.id,
            DeploymentUpdate.from_operational_state(operational_state),
        )

    def _check_deployment_inputs_outputs(
        self,
        snapshot: PipelineSnapshotResponse,
    ) -> None:
        """Check if the deployment has compiled schemas for the pipeline inputs and outputs.

        Args:
            snapshot: The pipeline snapshot to check.

        Raises:
            DeploymentProvisionError: if the deployment has no compiled schemas
                for the pipeline inputs and outputs.
        """
        if (
            not snapshot.pipeline_spec
            or not snapshot.pipeline_spec.input_schema
            or not snapshot.pipeline_spec.output_schema
        ):
            raise DeploymentProvisionError(
                f"The pipeline with name '{snapshot.pipeline.name}' referenced "
                f"by the deployment with name or ID "
                f"'{snapshot.name or snapshot.id}' "
                "is missing the compiled schemas for the pipeline inputs or "
                "outputs. This is most likely because some of the pipeline "
                "inputs or outputs are not JSON serializable. Please check that "
                "all the pipeline input arguments and return values have data "
                "types that are JSON serializable."
            )

    def _check_deployment_deployer(
        self,
        deployment: DeploymentResponse,
    ) -> None:
        """Check if the deployment is managed by this deployer.

        Args:
            deployment: The deployment to check.

        Raises:
            DeploymentDeployerMismatchError: if the deployment is
                not managed by this deployer.
        """
        if deployment.deployer_id and deployment.deployer_id != self.id:
            deployer = deployment.deployer
            assert deployer, "Deployer not found"
            raise DeploymentDeployerMismatchError(
                f"The deployment with name '{deployment.name}' was provisioned "
                f"with a deployer stack component ({deployer.name}) that is "
                f"different from the active one: {self.name}. "
                f"You can try one of the following:\n"
                f"1. Use a different name for the deployment\n"
                f"2. Delete the existing '{deployment.name}' deployment\n"
                f"3. Use a stack that contains the '{self.name}' deployer stack "
                "component\n"
            )

    def _check_deployment_snapshot(
        self, snapshot: Optional[PipelineSnapshotResponse] = None
    ) -> None:
        """Check if the snapshot was created for this deployer.

        Args:
            snapshot: The pipeline snapshot to check.

        Raises:
            DeploymentSnapshotMismatchError: if the pipeline snapshot is
                not built for this deployer.
        """
        if not snapshot:
            return

        if snapshot.stack and snapshot.stack.components.get(
            StackComponentType.DEPLOYER
        ):
            deployer = snapshot.stack.components[StackComponentType.DEPLOYER][
                0
            ]
            if deployer.id != self.id:
                raise DeploymentSnapshotMismatchError(
                    f"The pipeline snapshot with ID '{snapshot.id}' "
                    f"was not created for the deployer {self.name}. This will "
                    "lead to unexpected behavior and is not allowed."
                )

    def _check_snapshot_already_deployed(
        self,
        snapshot: PipelineSnapshotResponse,
        new_deployment_id_or_name: Union[str, UUID],
    ) -> None:
        """Check if the snapshot is already deployed to another deployment.

        Args:
            snapshot: The pipeline snapshot to check.
            new_deployment_id_or_name: The ID or name of the deployment that is
                being provisioned.

        Raises:
            DeploymentAlreadyExistsError: if the snapshot is already deployed to
                another deployment.
        """
        if snapshot.deployment and (
            isinstance(snapshot.deployment.id, UUID)
            and snapshot.deployment.id != new_deployment_id_or_name
            or (
                isinstance(snapshot.deployment.id, str)
                and snapshot.deployment.name != new_deployment_id_or_name
            )
        ):
            raise DeploymentAlreadyExistsError(
                f"The pipeline snapshot with name or ID "
                f"'{snapshot.name or snapshot.id}' "
                f"already has an associated deployment: "
                f"'{snapshot.deployment.name or snapshot.deployment.id}'. "
                "You can try one of the following:\n"
                "1. Delete the existing deployment before provisioning "
                f"a new one: 'zenml deployment delete "
                f"{snapshot.deployment.name or snapshot.deployment.id}'\n"
                "2. Update the existing deployment with the snapshot: 'zenml "
                f"pipeline snapshot deploy {snapshot.name or snapshot.id} "
                f"--deployment {snapshot.deployment.name or snapshot.deployment.id}'\n"
                "3. Create and deploy a different snapshot: 'zenml snapshot "
                "create ...'\n"
            )

    def _generate_auth_key(self, key_length: int = 32) -> str:
        """Generate an authentication key.

        Args:
            key_length: The length of the authentication key.

        Returns:
            The generated authentication key.
        """
        # Generate a secure random string with letters, digits and special chars
        alphabet = string.ascii_letters + string.digits
        return "".join(secrets.choice(alphabet) for _ in range(key_length))

    def _check_deployment_health(
        self,
        deployment: DeploymentResponse,
    ) -> bool:
        """Check if the deployment is healthy by calling its health check endpoint.

        Args:
            deployment: The deployment to check.

        Returns:
            True if the deployment is healthy, False otherwise.
        """
        assert deployment.snapshot, "Deployment snapshot not found"

        settings = (
            deployment.snapshot.pipeline_configuration.deployment_settings
        )

        # If the health check endpoint is disabled, we consider the deployment healthy.
        if (
            DeploymentDefaultEndpoints.HEALTH
            not in settings.include_default_endpoints
        ):
            return True

        if not deployment.url:
            return False

        health_check_path = f"{settings.root_url_path}{settings.api_url_path}{settings.health_url_path}"
        health_check_url = f"{deployment.url}{health_check_path}"

        # Attempt to connect to the deployment and check if it is healthy
        try:
            response = requests.get(health_check_url, timeout=3)
            if response.status_code == 200:
                return True
            else:
                logger.debug(
                    f"Health check endpoint for deployment '{deployment.name}' "
                    f"at '{health_check_url}' returned status code "
                    f"{response.status_code}"
                )
                return False
        except Exception as e:
            logger.debug(
                f"Health check endpoint for deployment '{deployment.name}' "
                f"at '{health_check_url}' is not reachable: {e}"
            )
            return False

    def _poll_deployment(
        self,
        deployment: DeploymentResponse,
        desired_status: DeploymentStatus,
        timeout: int,
    ) -> Tuple[DeploymentResponse, DeploymentOperationalState]:
        """Poll the deployment until it reaches the desired status, an error occurs or times out.

        Args:
            deployment: The deployment to poll.
            desired_status: The desired status of the deployment.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to reach the desired status.

        Returns:
            The updated deployment and the operational state of the
            deployment.

        Raises:
            DeploymentTimeoutError: if the deployment
                deployment times out while waiting to reach the desired status.
        """
        logger.info(
            f"Waiting for the deployment {deployment.name} to reach "
            f"desired state '{desired_status}' for max {timeout} seconds..."
        )

        start_time = time.time()
        sleep_time = 5
        while True:
            deployment_state = DeploymentOperationalState(
                status=DeploymentStatus.ERROR,
            )
            try:
                deployment_state = self.do_get_deployment_state(deployment)

                if deployment_state.status == DeploymentStatus.RUNNING:
                    if not self._check_deployment_health(deployment):
                        deployment_state.status = DeploymentStatus.PENDING

            except DeploymentNotFoundError:
                deployment_state = DeploymentOperationalState(
                    status=DeploymentStatus.ABSENT
                )
            except DeployerError as e:
                logger.exception(
                    f"Failed to get deployment {deployment.name}: {e}"
                )
            finally:
                deployment = self._update_deployment(
                    deployment, deployment_state
                )

            if deployment.status in [
                desired_status,
                DeploymentStatus.ERROR,
            ]:
                break

            elapsed_time = int(time.time() - start_time)
            if elapsed_time > timeout:
                raise DeploymentTimeoutError(
                    f"Timed out waiting for deployment {deployment.name} "
                    f"to reach desired state '{desired_status}' after {timeout} "
                    "seconds"
                )
            logger.debug(
                f"The deployment {deployment.name} state is still "
                f"'{deployment.status}' after {elapsed_time} seconds. Waiting for "
                f"max {timeout - elapsed_time} seconds..."
            )
            time.sleep(sleep_time)

        return deployment, deployment_state

    def _get_deployment_analytics_metadata(
        self,
        deployment: "DeploymentResponse",
        stack: Optional["Stack"] = None,
    ) -> Dict[str, Any]:
        """Returns the deployment metadata.

        Args:
            deployment: The deployment to track.
            stack: The stack on which the pipeline is deployed.

        Returns:
            the metadata about the deployment
        """
        snapshot = deployment.snapshot
        stack_metadata = {}
        if stack:
            stack_metadata = {
                component_type.value: component.flavor
                for component_type, component in stack.components.items()
            }
        return {
            "project_id": deployment.project_id,
            "store_type": Client().zen_store.type.value,
            **stack_metadata,
            "deployment_id": str(deployment.id),
            "snapshot_id": str(snapshot.id) if snapshot else None,
            "deployer_id": str(self.id),
            "deployer_flavor": self.flavor,
            "deployment_status": deployment.status,
        }

    def provision_deployment(
        self,
        snapshot: PipelineSnapshotResponse,
        stack: "Stack",
        deployment_name_or_id: Union[str, UUID],
        replace: bool = True,
        timeout: Optional[int] = None,
    ) -> DeploymentResponse:
        """Provision a deployment.

        The provision_deployment method is the main entry point for
        provisioning deployments using the deployer. It is used to deploy
        a pipeline snapshot as an HTTP deployment, or update an existing
        deployment instance with the same name. The method returns a
        DeploymentResponse object that is a representation of the
        external deployment instance.

        Args:
            snapshot: The pipeline snapshot to deploy as an HTTP deployment.
            stack: The stack the pipeline will be deployed on.
            deployment_name_or_id: Unique name or ID for the deployment.
                This name must be unique at the project level.
            replace: If True, it will update in-place any existing pipeline
                deployment instance with the same name. If False, and the pipeline
                deployment instance already exists, it will raise a
                DeploymentAlreadyExistsError.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be provisioned. If provided, will override the
                deployer's default timeout.

        Raises:
            DeploymentAlreadyExistsError: if the deployment already
                exists and replace is False.
            DeploymentProvisionError: if the deployment fails.
            DeploymentSnapshotMismatchError: if the pipeline snapshot
                was not created for this deployer.
            DeploymentNotFoundError: if the deployment with the
                given ID is not found.
            DeployerError: if an unexpected error occurs.

        Returns:
            The DeploymentResponse object representing the provisioned
            deployment.
        """
        if not replace and is_valid_uuid(deployment_name_or_id):
            raise DeploymentAlreadyExistsError(
                f"A deployment with ID '{deployment_name_or_id}' "
                "already exists"
            )

        self._check_deployment_inputs_outputs(snapshot)

        client = Client()

        settings = cast(
            BaseDeployerSettings,
            self.get_settings(snapshot),
        )

        timeout = timeout or settings.lcm_timeout
        auth_key = settings.auth_key
        if not auth_key and settings.generate_auth_key:
            auth_key = self._generate_auth_key()

        if snapshot.stack and snapshot.stack.id != stack.id:
            # When a different stack is used then the one the snapshot was
            # created for, the container image may not have the correct
            # dependencies installed, which leads to unexpected errors during
            # deployment. To avoid this, we raise an error here.
            raise DeploymentSnapshotMismatchError(
                f"The pipeline snapshot with ID '{snapshot.id}' "
                f"was not created for the stack {stack.name} and might not "
                "have the correct dependencies installed. This may "
                "lead to unexpected behavior during deployment. Please switch "
                f"to the correct active stack '{snapshot.stack.name}' or use "
                "a different snapshot."
            )

        try:
            # Get the existing deployment
            deployment = client.get_deployment(
                deployment_name_or_id, project=snapshot.project_id
            )

            self._check_snapshot_already_deployed(snapshot, deployment.id)

            logger.debug(
                f"Existing deployment found with name '{deployment.name}'"
            )
        except KeyError:
            if isinstance(deployment_name_or_id, UUID):
                raise DeploymentNotFoundError(
                    f"Deployment with ID '{deployment_name_or_id}' not found"
                )

            self._check_snapshot_already_deployed(
                snapshot, deployment_name_or_id
            )

            logger.debug(
                f"Creating new deployment {deployment_name_or_id} with "
                f"snapshot ID: {snapshot.id}"
            )

            # Create the deployment request
            deployment_request = DeploymentRequest(
                name=deployment_name_or_id,
                project=snapshot.project_id,
                snapshot_id=snapshot.id,
                deployer_id=self.id,  # This deployer's ID
                auth_key=auth_key,
            )

            deployment = client.zen_store.create_deployment(deployment_request)
            logger.debug(
                f"Created new deployment with name '{deployment.name}' "
                f"and ID: {deployment.id}"
            )
        else:
            if not replace:
                raise DeploymentAlreadyExistsError(
                    f"A deployment with name '{deployment.name}' "
                    "already exists"
                )

            self._check_deployment_deployer(deployment)
            self._check_deployment_snapshot(snapshot)

            deployment_update = DeploymentUpdate(
                snapshot_id=snapshot.id,
            )
            if (
                deployment.auth_key
                and not auth_key
                or not deployment.auth_key
                and auth_key
            ):
                # Key was either added or removed
                deployment_update.auth_key = auth_key
            elif deployment.auth_key != auth_key and (
                settings.auth_key or not settings.generate_auth_key
            ):
                # Key was changed and not because of re-generation
                deployment_update.auth_key = auth_key

            # The deployment has been updated
            deployment = client.zen_store.update_deployment(
                deployment.id,
                deployment_update,
            )

        logger.info(
            f"Provisioning deployment {deployment.name} with "
            f"snapshot ID: {snapshot.id}"
        )

        environment, secrets = get_config_environment_vars(
            deployment_id=deployment.id,
        )

        # Make sure to use the correct active stack/project which correspond
        # to the supplied stack and snapshot, which may be different from the
        # active stack/project
        environment[ENV_ZENML_ACTIVE_STACK_ID] = str(stack.id)
        environment[ENV_ZENML_ACTIVE_PROJECT_ID] = str(snapshot.project_id)

        start_time = time.time()
        deployment_state = DeploymentOperationalState(
            status=DeploymentStatus.ERROR,
        )
        with track_handler(
            AnalyticsEvent.DEPLOY_PIPELINE
        ) as analytics_handler:
            try:
                deployment_state = self.do_provision_deployment(
                    deployment,
                    stack=stack,
                    environment=environment,
                    secrets=secrets,
                    timeout=timeout,
                )
            except DeploymentProvisionError as e:
                raise DeploymentProvisionError(
                    f"Failed to provision deployment {deployment.name}: {e}"
                ) from e
            except DeployerError as e:
                raise DeployerError(
                    f"Failed to provision deployment {deployment.name}: {e}"
                ) from e
            except Exception as e:
                raise DeployerError(
                    f"Unexpected error while provisioning deployment for "
                    f"{deployment.name}: {e}"
                ) from e
            finally:
                deployment = self._update_deployment(
                    deployment, deployment_state
                )

            logger.info(
                f"Provisioned deployment {deployment.name} with "
                f"snapshot ID: {snapshot.id}. Operational state is: "
                f"{deployment_state.status}"
            )

            try:
                if deployment_state.status == DeploymentStatus.RUNNING:
                    return deployment

                # Subtract the time spent deploying the deployment from the
                # timeout
                timeout = timeout - int(time.time() - start_time)
                deployment, _ = self._poll_deployment(
                    deployment, DeploymentStatus.RUNNING, timeout
                )

                if deployment.status != DeploymentStatus.RUNNING:
                    raise DeploymentProvisionError(
                        f"Failed to provision deployment {deployment.name}: "
                        f"The deployment's operational state is "
                        f"{deployment.status}. Please check the status or logs "
                        "of the deployment for more information."
                    )

            finally:
                analytics_handler.metadata = (
                    self._get_deployment_analytics_metadata(
                        deployment=deployment,
                        stack=stack,
                    )
                )

            return deployment

    def refresh_deployment(
        self,
        deployment_name_or_id: Union[str, UUID],
        project: Optional[UUID] = None,
    ) -> DeploymentResponse:
        """Refresh the status of a deployment by name or ID.

        Call this to refresh the operational state of a deployment.

        Args:
            deployment_name_or_id: The name or ID of the deployment to get.
            project: The project ID of the deployment to get. Required
                if a name is provided.

        Returns:
            The deployment.

        Raises:
            DeploymentNotFoundError: if the deployment is not found.
            DeployerError: if an unexpected error occurs.
        """
        client = Client()
        try:
            deployment = client.get_deployment(
                deployment_name_or_id, project=project
            )
        except KeyError:
            raise DeploymentNotFoundError(
                f"Deployment with name or ID '{deployment_name_or_id}' "
                f"not found"
            )

        self._check_deployment_deployer(deployment)

        deployment_state = DeploymentOperationalState(
            status=DeploymentStatus.ERROR,
        )
        try:
            deployment_state = self.do_get_deployment_state(deployment)
            if deployment_state.status == DeploymentStatus.RUNNING:
                if not self._check_deployment_health(deployment):
                    deployment_state.status = DeploymentStatus.PENDING
        except DeploymentNotFoundError:
            deployment_state.status = DeploymentStatus.ABSENT
        except DeployerError as e:
            raise DeployerError(
                f"Failed to refresh deployment {deployment_name_or_id}: {e}"
            ) from e
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while refreshing deployment for "
                f"{deployment_name_or_id}: {e}"
            ) from e
        finally:
            deployment = self._update_deployment(deployment, deployment_state)

        return deployment

    def deprovision_deployment(
        self,
        deployment_name_or_id: Union[str, UUID],
        project: Optional[UUID] = None,
        timeout: Optional[int] = None,
    ) -> DeploymentResponse:
        """Deprovision a deployment.

        Args:
            deployment_name_or_id: The name or ID of the deployment to
                deprovision.
            project: The project ID of the deployment to deprovision.
                Required if a name is provided.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to deprovision. If provided, will override the
                deployer's default timeout.

        Returns:
            The deployment.

        Raises:
            DeploymentNotFoundError: if the deployment is not found
                or is not managed by this deployer.
            DeploymentDeprovisionError: if the deployment
                deprovision fails.
            DeployerError: if an unexpected error occurs.
        """
        client = Client()
        try:
            deployment = client.get_deployment(
                deployment_name_or_id, project=project
            )
        except KeyError:
            raise DeploymentNotFoundError(
                f"Deployment with name or ID '{deployment_name_or_id}' "
                f"not found"
            )

        self._check_deployment_deployer(deployment)

        if not timeout and deployment.snapshot:
            settings = cast(
                BaseDeployerSettings,
                self.get_settings(deployment.snapshot),
            )

            timeout = settings.lcm_timeout

        timeout = timeout or DEFAULT_DEPLOYMENT_LCM_TIMEOUT

        start_time = time.time()
        deployment_state = DeploymentOperationalState(
            status=DeploymentStatus.ERROR,
        )
        with track_handler(
            AnalyticsEvent.STOP_DEPLOYMENT
        ) as analytics_handler:
            try:
                deleted_deployment_state = self.do_deprovision_deployment(
                    deployment, timeout
                )
                if not deleted_deployment_state:
                    # When do_delete_deployment returns a None value, this
                    # is to signal that the deployment is already fully deprovisioned.
                    deployment_state.status = DeploymentStatus.ABSENT
            except DeploymentNotFoundError:
                deployment_state.status = DeploymentStatus.ABSENT
            except DeployerError as e:
                raise DeployerError(
                    f"Failed to delete deployment {deployment_name_or_id}: {e}"
                ) from e
            except Exception as e:
                raise DeployerError(
                    f"Unexpected error while deleting deployment for "
                    f"{deployment_name_or_id}: {e}"
                ) from e
            finally:
                deployment = self._update_deployment(
                    deployment, deployment_state
                )

            try:
                if deployment_state.status == DeploymentStatus.ABSENT:
                    return deployment

                # Subtract the time spent deprovisioning the deployment from the timeout
                timeout = timeout - int(time.time() - start_time)
                deployment, _ = self._poll_deployment(
                    deployment, DeploymentStatus.ABSENT, timeout
                )

                if deployment.status != DeploymentStatus.ABSENT:
                    raise DeploymentDeprovisionError(
                        f"Failed to deprovision deployment {deployment_name_or_id}: "
                        f"Operational state: {deployment.status}"
                    )

            finally:
                analytics_handler.metadata = (
                    self._get_deployment_analytics_metadata(
                        deployment=deployment,
                        stack=None,
                    )
                )

            return deployment

    def delete_deployment(
        self,
        deployment_name_or_id: Union[str, UUID],
        project: Optional[UUID] = None,
        force: bool = False,
        timeout: Optional[int] = None,
    ) -> None:
        """Deprovision and delete a deployment.

        Args:
            deployment_name_or_id: The name or ID of the deployment to
                delete.
            project: The project ID of the deployment to deprovision.
                Required if a name is provided.
            force: if True, force the deployment to delete even if it
                cannot be deprovisioned.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be deprovisioned. If provided, will override the
                deployer's default timeout.

        Raises:
            DeployerError: if an unexpected error occurs.
        """
        client = Client()
        try:
            deployment = self.deprovision_deployment(
                deployment_name_or_id, project, timeout
            )
        except DeploymentNotFoundError:
            # The deployment was already deleted
            return
        except DeployerError as e:
            if force:
                logger.warning(
                    f"Failed to deprovision deployment "
                    f"{deployment_name_or_id}: {e}. Forcing deletion."
                )
                deployment = client.get_deployment(
                    deployment_name_or_id, project=project
                )
                client.zen_store.delete_deployment(deployment_id=deployment.id)
            else:
                raise
        else:
            client.zen_store.delete_deployment(deployment_id=deployment.id)

    def get_deployment_logs(
        self,
        deployment_name_or_id: Union[str, UUID],
        project: Optional[UUID] = None,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Get the logs of a deployment.

        Args:
            deployment_name_or_id: The name or ID of the deployment to get
                the logs of.
            project: The project ID of the deployment to get the logs of.
                Required if a name is provided.
            follow: if True, the logs will be streamed as they are written.
            tail: only retrieve the last NUM lines of log output.

        Returns:
            A generator that yields the logs of the deployment.

        Raises:
            DeploymentNotFoundError: if the deployment is not found.
            DeployerError: if an unexpected error occurs.
        """
        client = Client()
        try:
            deployment = client.get_deployment(
                deployment_name_or_id, project=project
            )
        except KeyError:
            raise DeploymentNotFoundError(
                f"Deployment with name or ID '{deployment_name_or_id}' "
                f"not found"
            )

        self._check_deployment_deployer(deployment)

        try:
            return self.do_get_deployment_state_logs(deployment, follow, tail)
        except DeployerError as e:
            raise DeployerError(
                f"Failed to get logs for deployment {deployment_name_or_id}: {e}"
            ) from e
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while getting logs for deployment for "
                f"{deployment_name_or_id}: {e}"
            ) from e

    # ------------------ Abstract Methods ------------------

    @abstractmethod
    def do_provision_deployment(
        self,
        deployment: DeploymentResponse,
        stack: "Stack",
        environment: Dict[str, str],
        secrets: Dict[str, str],
        timeout: int,
    ) -> DeploymentOperationalState:
        """Abstract method to deploy a pipeline as an HTTP deployment.

        Concrete deployer subclasses must implement the following
        functionality in this method:

        - Create the actual deployment infrastructure (e.g.,
        FastAPI server, Kubernetes deployment, cloud function, etc.) based on
        the information in the deployment response, particularly the
        pipeline snapshot. When determining how to name the external
        resources, do not rely on the deployment name as being immutable
        or unique.

        - If the deployment infrastructure is already provisioned, update
        it to match the information in the deployment response.

        - Return a DeploymentOperationalState representing the operational
        state of the provisioned deployment.

        Note that the deployment infrastructure is not required to be
        deployed immediately. The deployer can return a
        DeploymentOperationalState with a status of
        DeploymentStatus.PENDING, and the base deployer will poll
        the deployment infrastructure by calling the
        `do_get_deployment_state` method until it is ready or it times out.

        Args:
            deployment: The deployment to deploy as an HTTP deployment.
            stack: The stack the pipeline will be deployed on.
            environment: A dictionary of environment variables to set on the
                deployment.
            secrets: A dictionary of secret environment variables to set
                on the deployment. These secret environment variables
                should not be exposed as regular environment variables on the
                deployer.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be provisioned.

        Returns:
            The DeploymentOperationalState object representing the
            operational state of the provisioned deployment.

        Raises:
            DeploymentProvisionError: if provisioning the deployment
                fails.
            DeployerError: if an unexpected error occurs.
        """

    @abstractmethod
    def do_get_deployment_state(
        self,
        deployment: DeploymentResponse,
    ) -> DeploymentOperationalState:
        """Abstract method to get information about a deployment.

        Args:
            deployment: The deployment to get information about.

        Returns:
            The DeploymentOperationalState object representing the
            updated operational state of the deployment.

        Raises:
            DeploymentNotFoundError: if no deployment is found
                corresponding to the provided DeploymentResponse.
            DeployerError: if the deployment information cannot
                be retrieved for any other reason or if an unexpected error
                occurs.
        """

    @abstractmethod
    def do_get_deployment_state_logs(
        self,
        deployment: DeploymentResponse,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Abstract method to get the logs of a deployment.

        Args:
            deployment: The deployment to get the logs of.
            follow: if True, the logs will be streamed as they are written
            tail: only retrieve the last NUM lines of log output.

        Yields:
            The logs of the deployment.

        Raises:
            DeploymentNotFoundError: if no deployment is found
                corresponding to the provided DeploymentResponse.
            DeploymentLogsNotFoundError: if the deployment logs are not
                found.
            DeployerError: if the deployment logs cannot
                be retrieved for any other reason or if an unexpected error
                occurs.
        """

    @abstractmethod
    def do_deprovision_deployment(
        self,
        deployment: DeploymentResponse,
        timeout: int,
    ) -> Optional[DeploymentOperationalState]:
        """Abstract method to deprovision a deployment.

        Concrete deployer subclasses must implement the following
        functionality in this method:

        - Deprovision the actual deployment infrastructure (e.g.,
        FastAPI server, Kubernetes deployment, cloud function, etc.) based on
        the information in the deployment response.

        - Return a DeploymentOperationalState representing the operational
        state of the deleted deployment, or None if the deletion is
        completed before the call returns.

        Note that the deployment infrastructure is not required to be
        deleted immediately. The deployer can return a
        DeploymentOperationalState with a status of
        DeploymentStatus.PENDING, and the base deployer will poll
        the deployment infrastructure by calling the
        `do_get_deployment_state` method until it is deleted or it times out.

        Args:
            deployment: The deployment to delete.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be deprovisioned.

        Returns:
            The DeploymentOperationalState object representing the
            operational state of the deprovisioned deployment, or None
            if the deprovision is completed before the call returns.

        Raises:
            DeploymentNotFoundError: if no deployment is found
                corresponding to the provided DeploymentResponse.
            DeploymentDeprovisionError: if the deployment
                deprovision fails.
            DeployerError: if an unexpected error occurs.
        """


class BaseDeployerFlavor(Flavor):
    """Base class for deployer flavors."""

    @property
    def type(self) -> StackComponentType:
        """Returns the flavor type.

        Returns:
            The flavor type.
        """
        return StackComponentType.DEPLOYER

    @property
    def config_class(self) -> Type[BaseDeployerConfig]:
        """Returns `BaseDeployerConfig` config class.

        Returns:
                The config class.
        """
        return BaseDeployerConfig

    @property
    @abstractmethod
    def implementation_class(self) -> Type[BaseDeployer]:
        """The class that implements the deployer."""



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Request management utilities."""

import asyncio
import base64
import json
import random
import time
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional
from uuid import UUID, uuid4

from fastapi import Request, Response
from fastapi.responses import JSONResponse

from zenml.constants import MEDIUMTEXT_MAX_LENGTH
from zenml.exceptions import EntityExistsError
from zenml.logger import get_logger
from zenml.models import ApiTransactionRequest, ApiTransactionUpdate
from zenml.utils.json_utils import pydantic_encoder
from zenml.utils.time_utils import utc_now

if TYPE_CHECKING:
    from zenml.zen_server.auth import AuthContext


logger = get_logger(__name__)


class RequestContext:
    """Context for a request."""

    def __init__(self, request: Request) -> None:
        """Initialize the request context.

        Args:
            request: The request.

        Raises:
            ValueError: If the idempotency key is not a valid UUID.
        """
        self.request = request
        self.request_id = request.headers.get("X-Request-ID", str(uuid4())[:8])
        self.transaction_id: Optional[UUID] = None
        transaction_id = request.headers.get("Idempotency-Key")
        if transaction_id:
            try:
                self.transaction_id = UUID(transaction_id)
            except ValueError:
                raise ValueError(
                    f"Invalid UUID idempotency key: {transaction_id}. "
                    "Please use a valid UUID."
                )

        # Use a random trace ID to identify the request internally in the logs.
        self.trace_id = str(uuid4())[:4]

        self.source = request.headers.get("User-Agent") or ""
        self.received_at = utc_now()

        self.auth_context: Optional["AuthContext"] = None

    @property
    def process_time(self) -> float:
        """Get the process time in seconds of the request.

        Returns:
            The request ID.
        """
        return (utc_now() - self.received_at).total_seconds()

    @property
    def log_request_id(self) -> str:
        """Get the full request ID for logging.

        Returns:
            The request ID.
        """
        source_type = self.source.split("/")[0]
        return f"{self.request_id}/{source_type}/{self.trace_id}"

    @property
    def log_request(self) -> str:
        """Get the request details for logging.

        Returns:
            The request details for logging.
        """
        client_ip = (
            self.request.client.host if self.request.client else "unknown"
        )
        url_path = self.request.url.path
        method = self.request.method
        return f"{method} {url_path} from {client_ip}"

    @property
    def log_duration(self) -> str:
        """Get the duration of the request.

        Returns:
            The duration of the request.
        """
        current_time = utc_now()
        duration = (current_time - self.received_at).total_seconds() * 1000
        return f"{duration:.2f}ms"

    @property
    def is_cacheable(self) -> bool:
        """Check if the request is cacheable.

        Returns:
            Whether the request is cacheable.
        """
        # Only cache requests that are authenticated and are part of a
        # transaction.
        return (
            self.auth_context is not None and self.transaction_id is not None
        )


class RequestRecord:
    """A record of an in-flight or cached request."""

    future: asyncio.Future[Any]
    request_context: RequestContext

    def __init__(
        self, future: asyncio.Future[Any], request_context: RequestContext
    ) -> None:
        """Initialize the request record.

        Args:
            future: The future of the request.
            request_context: The request context.
        """
        self.future = future
        self.request_context = request_context
        self.completed = False

    def set_result(self, result: Any) -> None:
        """Set the result of the request.

        Args:
            result: The result of the request.
        """
        self.future.set_result(result)
        self.completed = True

    def set_exception(self, exception: Exception) -> None:
        """Set the exception of the request.

        Args:
            exception: The exception of the request.
        """
        self.future.set_exception(exception)
        self.completed = True


class RequestManager:
    """A manager for requests.

    This class is used to manage requests by caching the results of requests
    that have already been executed.
    """

    def __init__(
        self,
    ) -> None:
        """Initialize the request manager."""
        from zenml.zen_server.utils import server_config

        cfg = server_config()
        self.deduplicate = cfg.request_deduplication
        self.transactions: Dict[UUID, RequestRecord] = dict()
        self.lock = asyncio.Lock()
        self.transaction_ttl = cfg.request_cache_timeout
        self.request_timeout = cfg.request_timeout

        self.request_contexts: ContextVar[Optional[RequestContext]] = (
            ContextVar("request_contexts", default=None)
        )

        self._cleanup_interval = cfg.api_transaction_cleanup_interval

        self._cleanup_task: Optional[asyncio.Task[None]] = None
        self._shutdown_event = asyncio.Event()

    @property
    def current_request(self) -> RequestContext:
        """Get the current request context.

        Returns:
            The current request context.

        Raises:
            RuntimeError: If no request context is set.
        """
        request_context = self.request_contexts.get()
        if request_context is None:
            raise RuntimeError("No request context set")
        return request_context

    @current_request.setter
    def current_request(self, request_context: RequestContext) -> None:
        """Set the current request context.

        Args:
            request_context: The request context.
        """
        self.request_contexts.set(request_context)

    async def startup(self) -> None:
        """Start the request manager."""
        self._shutdown_event.clear()
        self._cleanup_task = asyncio.create_task(self._run_cleanup_loop())

    async def shutdown(self) -> None:
        """Shutdown the request manager."""
        self._shutdown_event.set()
        if self._cleanup_task is not None:
            self._cleanup_task.cancel()
            try:
                await self._cleanup_task
            except asyncio.CancelledError:
                pass
            self._cleanup_task = None

    async def _run_cleanup_loop(self) -> None:
        """Run the cleanup loop for expired transactions.

        This loop runs continuously in the background, cleaning up expired
        API transactions. It includes:
        - Random jitter on startup and between iterations to reduce overlap
        between pods
        - Graceful shutdown via the shutdown event
        """
        # Random initial delay (0-interval seconds) to stagger pod startup
        initial_jitter = random.uniform(0, self._cleanup_interval)
        logger.debug(
            f"Transaction cleanup will start in {initial_jitter:.1f}s"
        )
        try:
            await asyncio.wait_for(
                self._shutdown_event.wait(), timeout=initial_jitter
            )
            return
        except asyncio.TimeoutError:
            pass

        while not self._shutdown_event.is_set():
            try:
                await asyncio.get_event_loop().run_in_executor(
                    None, self._cleanup_expired_transactions
                )
            except Exception:
                logger.exception("Error during transaction cleanup")

            # Add random jitter (±25%) to interval to reduce pod overlap
            jitter = random.uniform(0.75, 1.25)
            sleep_time = self._cleanup_interval * jitter

            try:
                await asyncio.wait_for(
                    self._shutdown_event.wait(), timeout=sleep_time
                )
                break
            except asyncio.TimeoutError:
                pass

    def _cleanup_expired_transactions(self) -> None:
        """Execute the cleanup operation in a thread pool."""
        from zenml.zen_server.utils import zen_store

        start_time = time.perf_counter()
        zen_store().cleanup_expired_api_transactions()
        duration = time.perf_counter() - start_time

        logger.debug(f"Cleaning up expired transactions took {duration:.2f}s")

    async def async_run_and_cache_result(
        self,
        func: Callable[..., Any],
        deduplicate: bool,
        request_record: RequestRecord,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Run a request and cache the result.

        This method is called in the background to run a request and cache the
        result.

        Args:
            func: The function to execute.
            deduplicate: Whether to enable or disable request deduplication for
                this request.
            request_record: The request record to cache the result in.
            *args: The arguments to pass to the function.
            **kwargs: The keyword arguments to pass to the function.
        """
        from starlette.concurrency import run_in_threadpool

        from zenml.zen_server.utils import get_system_metrics_log_str

        request_context = request_record.request_context
        transaction_id = request_context.transaction_id

        logger.debug(
            f"[{request_context.log_request_id}] ENDPOINT STATS - "
            f"{request_context.log_request} "
            f"async {func.__name__} STARTED "
            f"{get_system_metrics_log_str(request_context.request)}"
        )

        def sync_run_and_cache_result(*args: Any, **kwargs: Any) -> Any:
            from zenml.zen_server.utils import zen_store

            # Copy the deduplicate flag to a local variable to avoid modifying
            # the argument in the outer scope.
            deduplicate_request = deduplicate

            logger.debug(
                f"[{request_context.log_request_id}] ENDPOINT STATS - "
                f"{request_context.log_request} "
                f"sync {func.__name__} STARTED "
                f"{get_system_metrics_log_str(request_context.request)}"
            )

            try:
                # Create or get the API transaction from the database
                if deduplicate_request:
                    assert transaction_id is not None
                    try:
                        api_transaction, transaction_created = (
                            zen_store().get_or_create_api_transaction(
                                api_transaction=ApiTransactionRequest(
                                    transaction_id=transaction_id,
                                    method=request_context.request.method,
                                    url=str(request_context.request.url),
                                )
                            )
                        )
                    except EntityExistsError:
                        logger.error(
                            f"[{request_context.log_request_id}] "
                            f"Transaction {transaction_id} already exists "
                            f"with method {request_context.request.method} and "
                            f"URL {str(request_context.request.url)}. Skipping "
                            "caching."
                        )
                        deduplicate_request = False
                    except Exception:
                        logger.exception(
                            f"[{request_context.log_request_id}] "
                            f"Unexpected error getting or creating API "
                            "transaction. Skipping deduplication."
                        )
                        deduplicate_request = False

                    if deduplicate_request:
                        if api_transaction.completed:
                            logger.debug(
                                f"[{request_context.log_request_id}] "
                                "ENDPOINT STATS - "
                                f"{request_context.log_request} "
                                f"sync {func.__name__} CACHE HIT "
                                f"{get_system_metrics_log_str(request_context.request)}"
                            )

                            try:
                                # The transaction already completed, we can return the
                                # result right away.
                                result = (
                                    zen_store().get_api_transaction_result(
                                        api_transaction_id=transaction_id,
                                    )
                                )
                            except Exception:
                                logger.exception(
                                    f"[{request_context.log_request_id}] "
                                    f"Unexpected error getting API transaction "
                                    "result. Skipping deduplication."
                                )
                                deduplicate_request = False
                            else:
                                if result is not None:
                                    return Response(
                                        base64.b64decode(result),
                                        media_type="application/json",
                                    )
                                else:
                                    return

                        elif not transaction_created:
                            logger.debug(
                                f"[{request_context.log_request_id}] "
                                "ENDPOINT STATS - "
                                f"{request_context.log_request} "
                                f"sync {func.__name__} DELAYED "
                                f"{get_system_metrics_log_str(request_context.request)}"
                            )

                            # The transaction is being processed by another server
                            # instance. We need to wait for it to complete. Instead
                            # of blocking this worker thread, we return a 429 error
                            # to the client to force it to retry later.
                            return JSONResponse(
                                {
                                    "error": "Server too busy. Please try again later."
                                },
                                status_code=429,
                            )

                try:
                    result = func(*args, **kwargs)
                except Exception:
                    if deduplicate_request:
                        assert transaction_id is not None
                        try:
                            # We don't cache exceptions. If the client retries, the
                            # request will be executed again and the exception, if
                            # persistent, will be raised again.
                            zen_store().delete_api_transaction(
                                api_transaction_id=transaction_id,
                            )
                        except Exception:
                            logger.exception(
                                f"[{request_context.log_request_id}] "
                                f"Unexpected error deleting API transaction."
                            )
                    raise

                if deduplicate_request:
                    assert transaction_id is not None
                    cache_result = True
                    result_to_cache: Optional[bytes] = None
                    if result is not None:
                        try:
                            result_to_cache = base64.b64encode(
                                json.dumps(
                                    result, default=pydantic_encoder
                                ).encode("utf-8")
                            )
                        except Exception:
                            # If the result is not serializable, we don't cache it.
                            cache_result = False
                            logger.exception(
                                f"Failed to serialize result of {func.__name__} "
                                f"for transaction {transaction_id}. Skipping "
                                "caching."
                            )
                        else:
                            if len(result_to_cache) > MEDIUMTEXT_MAX_LENGTH:
                                # If the result is too large, we also don't cache it.
                                cache_result = False
                                result_to_cache = None
                                logger.error(
                                    f"Result of {func.__name__} "
                                    f"for transaction {transaction_id} is too "
                                    "large. Skipping caching."
                                )

                    if cache_result:
                        api_transaction_update = ApiTransactionUpdate(
                            cache_time=self.transaction_ttl,
                        )
                        if result_to_cache is not None:
                            api_transaction_update.set_result(
                                result_to_cache.decode("utf-8")
                            )
                        try:
                            zen_store().finalize_api_transaction(
                                api_transaction_id=transaction_id,
                                api_transaction_update=api_transaction_update,
                            )
                        except Exception:
                            logger.exception(
                                f"[{request_context.log_request_id}] "
                                f"Unexpected error finalizing API transaction. "
                                "Ignoring."
                            )
                    else:
                        try:
                            # If the result is not cacheable, there is no point in
                            # keeping the transaction around.
                            zen_store().delete_api_transaction(
                                api_transaction_id=transaction_id,
                            )
                        except Exception:
                            logger.exception(
                                f"[{request_context.log_request_id}] "
                                f"Unexpected error deleting API transaction. "
                                "Ignoring."
                            )

                return result
            finally:
                logger.debug(
                    f"[{request_context.log_request_id}] ENDPOINT STATS - "
                    f"{request_context.log_request} "
                    f"sync {func.__name__} COMPLETED "
                    f"{get_system_metrics_log_str(request_context.request)}"
                )

        try:
            result = await run_in_threadpool(
                sync_run_and_cache_result, *args, **kwargs
            )
        except Exception as e:
            result = e

        if deduplicate:
            async with self.lock:
                if transaction_id in self.transactions:
                    del self.transactions[transaction_id]

        logger.debug(
            f"[{request_context.log_request_id}] ENDPOINT STATS - "
            f"{request_context.log_request} "
            f"async {func.__name__} COMPLETED "
            f"{get_system_metrics_log_str(request_context.request)}"
        )

        if isinstance(result, Exception):
            request_record.set_exception(result)
        else:
            request_record.set_result(result)

    async def execute(
        self,
        func: Callable[..., Any],
        deduplicate: Optional[bool],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Execute a request with in-memory de-duplication.

        Call this method to execute a request with in-memory de-duplication.
        If the request (identified by the request_id) has previously been
        executed and the result is still cached, the result will be returned
        immediately. If the request is in-flight, the method will wait for it
        to complete. If the request is not in-flight, the method will start
        execution in the background and cache the result.

        Args:
            func: The function to execute.
            deduplicate: Whether to enable or disable request deduplication for
                this request. If not specified, by default, the deduplication
                is enabled for POST requests and disabled for other requests.
            *args: The arguments to pass to the function.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            The result of the request.
        """
        from zenml.zen_server.utils import get_system_metrics_log_str

        request_context = self.current_request
        assert request_context is not None

        logger.debug(
            f"[{request_context.log_request_id}] ENDPOINT STATS - "
            f"{request_context.log_request} "
            f"{func.__name__} STARTED "
            f"{get_system_metrics_log_str(request_context.request)}"
        )

        transaction_id = request_context.transaction_id

        if deduplicate is None:
            # If not specified, by default, the deduplication is enabled for
            # POST requests and disabled for other requests.
            deduplicate = request_context.request.method == "POST"

        deduplicate = (
            deduplicate and self.deduplicate and request_context.is_cacheable
        )

        async with self.lock:
            if deduplicate and transaction_id in self.transactions:
                # The transaction is still being processed on the same
                # server instance. We just wait for it to complete.
                fut = self.transactions[transaction_id].future
                logger.debug(
                    f"[{request_context.log_request_id}] ENDPOINT STATS - "
                    f"{request_context.log_request} "
                    f"{func.__name__} RESUMED "
                    f"{get_system_metrics_log_str(request_context.request)}"
                )
            else:
                # Start execution in background, use the future to wait for it
                # to complete.
                fut = asyncio.get_event_loop().create_future()
                request_record = RequestRecord(
                    future=fut, request_context=request_context
                )
                if deduplicate:
                    assert transaction_id is not None
                    # Also record the transaction for deduplication.
                    self.transactions[transaction_id] = request_record
                asyncio.create_task(
                    self.async_run_and_cache_result(
                        func,
                        deduplicate,
                        request_record,
                        *args,
                        **kwargs,
                    )
                )

        # Wait for the request to complete; timeout if deduplication is enabled
        try:
            # We take into account the time that has already elapsed since the
            # request was received to avoid keeping the request for too long.
            timeout = max(
                0, self.request_timeout - request_context.process_time
            )
            result = await asyncio.wait_for(
                # We use asyncio.shield to prevent the request from being
                # cancelled when the timeout is reached.
                asyncio.shield(fut),
                timeout=timeout if deduplicate else None,
            )

            logger.debug(
                f"[{request_context.log_request_id}] ENDPOINT STATS - "
                f"{request_context.log_request} "
                f"{func.__name__} COMPLETED "
                f"{get_system_metrics_log_str(request_context.request)}"
            )
            return result
        except asyncio.TimeoutError:
            logger.debug(
                f"[{request_context.log_request_id}] ENDPOINT STATS - "
                f"{request_context.log_request} "
                f"{func.__name__} TIMEOUT "
                f"{get_system_metrics_log_str(request_context.request)}"
            )
            return JSONResponse(
                {"error": "Server too busy. Please try again later."},
                status_code=429,
            )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the GCP Cloud Run deployer."""

import math
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

from google.api_core import exceptions as google_exceptions
from google.cloud import run_v2, secretmanager
from google.cloud.logging_v2 import Client as LoggingClient
from google.protobuf.json_format import MessageToDict
from pydantic import BaseModel

from zenml.config.base_settings import BaseSettings
from zenml.config.resource_settings import ResourceSettings
from zenml.deployers.containerized_deployer import ContainerizedDeployer
from zenml.deployers.exceptions import (
    DeployerError,
    DeploymentDeprovisionError,
    DeploymentLogsNotFoundError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
)
from zenml.deployers.server.entrypoint_configuration import (
    DEPLOYMENT_ID_OPTION,
    DeploymentEntrypointConfiguration,
)
from zenml.enums import DeploymentStatus, StackComponentType
from zenml.integrations.gcp.flavors.gcp_deployer_flavor import (
    GCPDeployerConfig,
    GCPDeployerSettings,
)
from zenml.integrations.gcp.google_credentials_mixin import (
    GoogleCredentialsMixin,
)
from zenml.logger import get_logger
from zenml.models import (
    DeploymentOperationalState,
    DeploymentResponse,
)
from zenml.stack import StackValidator

if TYPE_CHECKING:
    from zenml.stack import Stack

logger = get_logger(__name__)

# Default resource and scaling configuration constants
DEFAULT_CPU = "1"
DEFAULT_MEMORY = "2Gi"
DEFAULT_MIN_INSTANCES = 1
DEFAULT_MAX_INSTANCES = 100
DEFAULT_CONCURRENCY = 80

# GCP Cloud Run built-in limits
GCP_CLOUD_RUN_MAX_INSTANCES = 1000


class CloudRunDeploymentMetadata(BaseModel):
    """Metadata for a Cloud Run deployment."""

    service_name: Optional[str] = None
    service_url: Optional[str] = None
    project_id: Optional[str] = None
    location: Optional[str] = None
    revision_name: Optional[str] = None
    reconciling: Optional[bool] = None
    service_status: Optional[Dict[str, Any]] = None
    cpu: Optional[str] = None
    memory: Optional[str] = None
    min_instances: Optional[int] = None
    max_instances: Optional[int] = None
    concurrency: Optional[int] = None
    timeout_seconds: Optional[int] = None
    ingress: Optional[str] = None
    vpc_connector: Optional[str] = None
    service_account: Optional[str] = None
    execution_environment: Optional[str] = None
    port: Optional[int] = None
    allow_unauthenticated: Optional[bool] = None
    labels: Optional[Dict[str, str]] = None
    annotations: Optional[Dict[str, str]] = None
    traffic_allocation: Optional[Dict[str, int]] = None
    created_time: Optional[str] = None
    updated_time: Optional[str] = None
    secrets: List[str] = []

    @classmethod
    def from_cloud_run_service(
        cls,
        service: run_v2.Service,
        project_id: str,
        location: str,
        secrets: List[secretmanager.Secret],
    ) -> "CloudRunDeploymentMetadata":
        """Create metadata from a Cloud Run service.

        Args:
            service: The Cloud Run service object.
            project_id: The GCP project ID.
            location: The GCP location.
            secrets: The list of existing GCP Secret Manager secrets for the
                deployment.

        Returns:
            The metadata for the Cloud Run service.
        """
        container = None
        if service.template and service.template.containers:
            container = service.template.containers[0]

        cpu = None
        memory = None
        if container and container.resources and container.resources.limits:
            cpu = container.resources.limits.get("cpu")
            memory = container.resources.limits.get("memory")

        min_instances = None
        max_instances = None
        if service.template and service.template.scaling:
            scaling = service.template.scaling
            min_instances = scaling.min_instance_count
            max_instances = scaling.max_instance_count

        concurrency = None
        if service.template:
            concurrency = service.template.max_instance_request_concurrency

        timeout_seconds = None
        if service.template and service.template.timeout:
            timeout_seconds = service.template.timeout.seconds

        ingress = None
        if service.ingress:
            ingress = str(service.ingress)

        vpc_connector = None
        if service.template and service.template.vpc_access:
            vpc_connector = service.template.vpc_access.connector

        service_account = None
        if service.template:
            service_account = service.template.service_account

        execution_environment = None
        if service.template and service.template.execution_environment:
            execution_environment = str(service.template.execution_environment)

        port = None
        if container and container.ports:
            port = container.ports[0].container_port

        traffic_allocation = {}
        if service.traffic:
            for traffic in service.traffic:
                if traffic.revision:
                    traffic_allocation[traffic.revision] = traffic.percent
                elif traffic.tag:
                    traffic_allocation[traffic.tag] = traffic.percent
                else:
                    traffic_allocation["LATEST"] = traffic.percent

        return cls(
            service_name=service.name.split("/")[-1] if service.name else None,
            service_url=service.uri if hasattr(service, "uri") else None,
            project_id=project_id,
            location=location,
            revision_name=(
                service.template.revision
                if service.template and service.template.revision
                else None
            ),
            reconciling=service.reconciling,
            service_status=MessageToDict(
                service.terminal_condition._pb,
            )
            if service.terminal_condition
            else None,
            cpu=cpu,
            memory=memory,
            min_instances=min_instances,
            max_instances=max_instances,
            concurrency=concurrency,
            timeout_seconds=timeout_seconds,
            ingress=ingress,
            vpc_connector=vpc_connector,
            service_account=service_account,
            execution_environment=execution_environment,
            port=port,
            allow_unauthenticated=True,
            labels=dict(service.labels) if service.labels else {},
            annotations=dict(service.annotations)
            if service.annotations
            else {},
            traffic_allocation=traffic_allocation,
            created_time=(
                service.create_time.isoformat()
                if service.create_time
                else None
            ),
            updated_time=(
                service.update_time.isoformat()
                if service.update_time
                else None
            ),
            secrets=[secret.name for secret in secrets],
        )

    @classmethod
    def from_deployment(
        cls, deployment: DeploymentResponse
    ) -> "CloudRunDeploymentMetadata":
        """Create metadata from a deployment.

        Args:
            deployment: The deployment to get the metadata for.

        Returns:
            The metadata for the deployment.
        """
        return cls.model_validate(deployment.deployment_metadata)


class GCPDeployer(ContainerizedDeployer, GoogleCredentialsMixin):
    """Deployer responsible for deploying pipelines on GCP Cloud Run."""

    _credentials: Optional[Any] = None
    _project_id: Optional[str] = None
    _cloud_run_client: Optional[run_v2.ServicesClient] = None
    _logging_client: Optional[LoggingClient] = None
    _secret_manager_client: Optional[
        secretmanager.SecretManagerServiceClient
    ] = None

    @property
    def config(self) -> GCPDeployerConfig:
        """Returns the `GCPDeployerConfig` config.

        Returns:
            The configuration.
        """
        return cast(GCPDeployerConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the GCP deployer.

        Returns:
            The settings class.
        """
        return GCPDeployerSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Ensures there is an image builder in the stack.

        Returns:
            A `StackValidator` instance.
        """
        return StackValidator(
            required_components={
                StackComponentType.IMAGE_BUILDER,
                StackComponentType.CONTAINER_REGISTRY,
            }
        )

    def _get_credentials_and_project_id(self) -> Tuple[Any, str]:
        """Get GCP credentials and project ID.

        Returns:
            A tuple containing the credentials and project ID.
        """
        if (
            self._credentials is not None
            and self._project_id is not None
            and not self.connector_has_expired()
        ):
            return self._credentials, self._project_id

        credentials, project_id = self._get_authentication()

        self._credentials = credentials
        self._project_id = project_id
        return credentials, project_id

    @property
    def project_id(self) -> str:
        """Get the GCP project ID.

        Returns:
            The GCP project ID.
        """
        _, project_id = self._get_credentials_and_project_id()
        return project_id

    @property
    def cloud_run_client(self) -> run_v2.ServicesClient:
        """Get the Cloud Run client.

        Returns:
            The Cloud Run client.
        """
        if self._cloud_run_client is None or self.connector_has_expired():
            credentials, _ = self._get_credentials_and_project_id()
            self._cloud_run_client = run_v2.ServicesClient(
                credentials=credentials
            )
        return self._cloud_run_client

    @property
    def logging_client(self) -> LoggingClient:
        """Get the Cloud Logging client.

        Returns:
            The Cloud Logging client.
        """
        if self._logging_client is None or self.connector_has_expired():
            credentials, project_id = self._get_credentials_and_project_id()
            self._logging_client = LoggingClient(
                project=project_id, credentials=credentials
            )
        return self._logging_client

    @property
    def secret_manager_client(
        self,
    ) -> secretmanager.SecretManagerServiceClient:
        """Get the Secret Manager client.

        Returns:
            The Secret Manager client.
        """
        if self._secret_manager_client is None or self.connector_has_expired():
            credentials, _ = self._get_credentials_and_project_id()
            self._secret_manager_client = (
                secretmanager.SecretManagerServiceClient(
                    credentials=credentials
                )
            )
        return self._secret_manager_client

    def get_labels(
        self, deployment: DeploymentResponse, settings: GCPDeployerSettings
    ) -> Dict[str, str]:
        """Get the labels for a deployment.

        Args:
            deployment: The deployment.
            settings: The deployer settings.

        Returns:
            The labels for the deployment.
        """
        return {
            **settings.labels,
            "zenml-deployment-id": str(deployment.id),
            "zenml-deployment-name": deployment.name,
            "zenml-deployer-name": str(self.name),
            "zenml-deployer-id": str(self.id),
            "managed-by": "zenml",
        }

    def _sanitize_name(
        self,
        name: str,
        random_suffix: str,
        max_length: int = 63,
    ) -> str:
        """Sanitize a name to comply with GCP naming requirements.

        Common GCP naming requirements:
        - Length: 1-max_length characters
        - Characters: lowercase letters (a-z), numbers (0-9), hyphens (-)
        - Must start and end with a letter or number

        Args:
            name: The raw name to sanitize.
            random_suffix: A random suffix to add to the name to ensure
                uniqueness.
            max_length: The maximum length of the name.

        Returns:
            A sanitized name that complies with GCP requirements.

        Raises:
            RuntimeError: If the random suffix is invalid.
            ValueError: If the service name is invalid.
        """
        if (
            not re.match(r"^[a-z0-9]+$", random_suffix)
            or len(random_suffix) < 1
        ):
            raise RuntimeError(
                f"Invalid random suffix: {random_suffix}. Must contain only "
                "lowercase letters and numbers and be at least 1 character "
                "long."
            )

        # Convert to lowercase and replace all disallowed characters with
        # hyphens
        sanitized = re.sub(r"[^a-z0-9-]", "-", name.lower())

        # Remove consecutive hyphens
        sanitized = re.sub(r"-+", "-", sanitized)

        # Remove leading and trailing hyphens before truncating
        sanitized = re.sub(
            r"^[^a-zA-Z0-9]+|[^a-zA-Z0-9]+$",
            "",
            sanitized,
        )

        # Truncate to fit within max_length character limit including suffix
        max_base_length = (
            max_length - len(random_suffix) - 1  # -1 for the hyphen
        )
        if len(sanitized) > max_base_length:
            sanitized = sanitized[:max_base_length]

        # Ensure it starts and ends with alphanumeric characters
        sanitized = re.sub(
            r"^[^a-zA-Z0-9]+|[^a-zA-Z0-9]+$",
            "",
            sanitized,
        )

        # Ensure we have at least one character after cleanup
        if not sanitized:
            raise ValueError(
                f"Invalid name: {name}. Must contain at least one "
                "alphanumeric character."
            )

        return f"{sanitized}-{random_suffix}"

    def _get_service_name(
        self, deployment_name: str, deployment_id: UUID, prefix: str
    ) -> str:
        """Get the Cloud Run service name for a deployment.

        Args:
            deployment_id: The deployment ID.
            deployment_name: The deployment name.
            prefix: The prefix to use for the service name.

        Returns:
            The Cloud Run service name that complies with all naming requirements.
        """
        deployment_id_short = str(deployment_id)[:8]
        raw_name = f"{prefix}{deployment_name}"

        return self._sanitize_name(
            raw_name, deployment_id_short, max_length=49
        )

    def _get_secret_name(
        self,
        deployment_id: UUID,
        env_var_name: str,
        prefix: str,
    ) -> str:
        """Get the Secret Manager secret name for an environment variable.

        Args:
            deployment_id: The deployment ID.
            env_var_name: The environment variable name.
            prefix: The prefix to use for the secret name.

        Returns:
            The Secret Manager secret name.
        """
        deployment_id_short = str(deployment_id)[:8]
        raw_name = f"{prefix}{env_var_name}"

        return self._sanitize_name(
            raw_name, deployment_id_short, max_length=255
        )

    def _create_or_update_secret(
        self,
        secret_name: str,
        secret_value: str,
        project_id: str,
        deployment: DeploymentResponse,
        settings: GCPDeployerSettings,
    ) -> secretmanager.Secret:
        """Create or update a secret in Secret Manager.

        Args:
            secret_name: The name of the secret.
            secret_value: The value to store.
            project_id: The GCP project ID.
            deployment: The deployment.
            settings: The deployer settings.

        Returns:
            The full secret.

        Raises:
            DeployerError: If secret creation/update fails.
        """
        parent = f"projects/{project_id}"
        secret_id = secret_name
        secret_path = f"{parent}/secrets/{secret_id}"

        try:
            try:
                secret = self.secret_manager_client.get_secret(
                    name=secret_path
                )
                logger.debug(
                    f"Secret {secret_name} already exists, adding new version"
                )
            except google_exceptions.NotFound:
                logger.debug(f"Creating new secret {secret_name}")
                secret = secretmanager.Secret(
                    replication=secretmanager.Replication(
                        automatic=secretmanager.Replication.Automatic()
                    ),
                    labels=self.get_labels(deployment, settings),
                )
                secret = self.secret_manager_client.create_secret(
                    parent=parent, secret_id=secret_id, secret=secret
                )

            payload = secretmanager.SecretPayload(
                data=secret_value.encode("utf-8")
            )
            version_response = self.secret_manager_client.add_secret_version(
                parent=secret_path, payload=payload
            )

            logger.debug(f"Created secret version: {version_response.name}")
            return secret

        except google_exceptions.GoogleAPICallError as e:
            raise DeployerError(
                f"Failed to create/update secret {secret_name}: {e}"
            )

    def _get_secrets(
        self, deployment: DeploymentResponse
    ) -> List[secretmanager.Secret]:
        """Get the existing GCP Secret Manager secrets for a deployment.

        Args:
            deployment: The deployment.

        Returns:
            The list of existing GCP Secret Manager secrets for the
            deployment.
        """
        metadata = CloudRunDeploymentMetadata.from_deployment(deployment)
        secrets: List[secretmanager.Secret] = []
        for secret_name in metadata.secrets:
            try:
                secret = self.secret_manager_client.get_secret(
                    name=secret_name
                )
                secrets.append(secret)
            except google_exceptions.NotFound:
                continue
            except google_exceptions.GoogleAPICallError:
                logger.exception(f"Failed to get secret {secret_name}")
                continue
        return secrets

    def _delete_secret(self, secret_name: str, project_id: str) -> None:
        """Delete a secret from Secret Manager.

        Args:
            secret_name: The name of the secret to delete.
            project_id: The GCP project ID.
        """
        secret_path = f"projects/{project_id}/secrets/{secret_name}"
        try:
            self.secret_manager_client.delete_secret(name=secret_path)
            logger.debug(f"Deleted secret {secret_path}")
        except google_exceptions.NotFound:
            logger.debug(f"Secret {secret_path} not found, skipping deletion")
        except google_exceptions.GoogleAPICallError:
            logger.exception(f"Failed to delete secret {secret_path}")

    def _cleanup_deployment_secrets(
        self,
        deployment: DeploymentResponse,
    ) -> None:
        """Clean up all secrets associated with a deployment.

        Args:
            deployment: The deployment.
        """
        secrets = self._get_secrets(deployment)

        for secret in secrets:
            _, project_id, _, secret_name = secret.name.split("/")
            self._delete_secret(secret_name, project_id)

    def _prepare_environment_variables(
        self,
        deployment: DeploymentResponse,
        environment: Dict[str, str],
        secrets: Dict[str, str],
        settings: GCPDeployerSettings,
        project_id: str,
    ) -> Tuple[List[run_v2.EnvVar], List[secretmanager.Secret]]:
        """Prepare environment variables for Cloud Run, handling secrets appropriately.

        Args:
            deployment: The deployment.
            environment: Regular environment variables.
            secrets: Sensitive environment variables.
            settings: The deployer settings.
            project_id: The GCP project ID.

        Returns:
            Tuple containing:
            - List of Cloud Run environment variable configurations.
            - List of active Secret Manager secrets.
        """
        env_vars = []

        merged_env = {**settings.environment_variables, **environment}
        for key, value in merged_env.items():
            env_vars.append(run_v2.EnvVar(name=key, value=value))

        active_secrets: List[secretmanager.Secret] = []
        if secrets:
            if settings.use_secret_manager:
                for key, value in secrets.items():
                    secret_name = self._get_secret_name(
                        deployment.id, key.lower(), settings.secret_name_prefix
                    )

                    try:
                        active_secret = self._create_or_update_secret(
                            secret_name,
                            value,
                            project_id,
                            deployment,
                            settings,
                        )

                        # Create environment variable that references the secret
                        env_var = run_v2.EnvVar(
                            name=key,
                            value_source=run_v2.EnvVarSource(
                                secret_key_ref=run_v2.SecretKeySelector(
                                    secret=secret_name, version="latest"
                                )
                            ),
                        )
                        env_vars.append(env_var)
                        active_secrets.append(active_secret)

                    except Exception as e:
                        logger.warning(
                            f"Failed to create secret for {key}, falling back "
                            f"to direct env var: {e}"
                        )
                        env_vars.append(run_v2.EnvVar(name=key, value=value))

                metadata = CloudRunDeploymentMetadata.from_deployment(
                    deployment
                )
                # Delete GCP secrets that are no longer needed
                active_secret_names = [
                    secret.name for secret in active_secrets
                ]
                for existing_secret_name in metadata.secrets:
                    if existing_secret_name not in active_secret_names:
                        _, project_id, _, secret_name = (
                            existing_secret_name.split("/")
                        )
                        self._delete_secret(secret_name, project_id)
            else:
                logger.warning(
                    "Storing secrets directly in environment variables. "
                    "Consider enabling use_secret_manager for better security."
                )
                for key, value in secrets.items():
                    env_vars.append(run_v2.EnvVar(name=key, value=value))

        return env_vars, active_secrets

    def _get_service_path(
        self,
        service_name: str,
        project_id: str,
        location: str,
    ) -> str:
        """Get the full Cloud Run service path.

        Args:
            service_name: The name of the Cloud Run service.
            project_id: The GCP project ID.
            location: The GCP location.

        Returns:
            The full Cloud Run service path.
        """
        return f"projects/{project_id}/locations/{location}/services/{service_name}"

    def _get_cloud_run_service(
        self, deployment: DeploymentResponse
    ) -> Optional[run_v2.Service]:
        """Get an existing Cloud Run service for a deployment.

        Args:
            deployment: The deployment.

        Returns:
            The Cloud Run service, or None if it doesn't exist.
        """
        existing_metadata = CloudRunDeploymentMetadata.from_deployment(
            deployment
        )

        if (
            not existing_metadata.service_name
            or not existing_metadata.location
            or not existing_metadata.project_id
        ):
            return None

        service_path = self._get_service_path(
            existing_metadata.service_name,
            existing_metadata.project_id,
            existing_metadata.location,
        )

        try:
            return self.cloud_run_client.get_service(name=service_path)
        except google_exceptions.NotFound:
            return None

    def _get_service_operational_state(
        self,
        service: run_v2.Service,
        project_id: str,
        location: str,
        secrets: List[secretmanager.Secret],
    ) -> DeploymentOperationalState:
        """Get the operational state of a Cloud Run service.

        Args:
            service: The Cloud Run service.
            project_id: The GCP project ID.
            location: The GCP location.
            secrets: The list of active Secret Manager secrets.

        Returns:
            The operational state of the Cloud Run service.
        """
        metadata = CloudRunDeploymentMetadata.from_cloud_run_service(
            service, project_id, location, secrets
        )

        state = DeploymentOperationalState(
            status=DeploymentStatus.UNKNOWN,
            metadata=metadata.model_dump(exclude_none=True),
        )

        # This flag is set while the service is being reconciled
        if service.reconciling:
            state.status = DeploymentStatus.PENDING
        else:
            if (
                service.terminal_condition.state
                == run_v2.Condition.State.CONDITION_SUCCEEDED
            ):
                state.status = DeploymentStatus.RUNNING
                state.url = service.uri
            elif (
                service.terminal_condition.state
                == run_v2.Condition.State.CONDITION_FAILED
            ):
                state.status = DeploymentStatus.ERROR
            elif service.terminal_condition.state in [
                run_v2.Condition.State.CONDITION_PENDING,
                run_v2.Condition.State.CONDITION_RECONCILING,
            ]:
                state.status = DeploymentStatus.PENDING
            else:
                state.status = DeploymentStatus.UNKNOWN

        return state

    def _convert_resource_settings_to_gcp_format(
        self,
        resource_settings: ResourceSettings,
    ) -> Tuple[str, str]:
        """Convert ResourceSettings to GCP Cloud Run resource format.

        GCP Cloud Run CPU constraints:
        - Fractional CPUs: 0.08 to < 1.0 (in increments of 0.01)
        - Integer CPUs: 1, 2, 4, 6, or 8 (no fractional values allowed >= 1.0)

        Args:
            resource_settings: The resource settings from pipeline configuration.

        Returns:
            Tuple of (cpu, memory) in GCP Cloud Run format.
        """
        cpu = DEFAULT_CPU
        if resource_settings.cpu_count is not None:
            cpu_count = resource_settings.cpu_count

            if cpu_count < 1.0:
                # For values < 1.0, allow fractional CPUs
                # Ensure minimum is 0.08 and round to 2 decimal places
                cpu_count = max(0.08, round(cpu_count, 2))
                cpu = str(cpu_count)
            else:
                # For values >= 1.0, round up to the nearest valid integer
                valid_cpu_values = [1, 2, 4, 6, 8]
                rounded_cpu = math.ceil(cpu_count)

                # Find the smallest valid CPU value that satisfies the requirement
                for valid_cpu in valid_cpu_values:
                    if valid_cpu >= rounded_cpu:
                        cpu = str(valid_cpu)
                        break
                else:
                    # If requested CPU exceeds maximum, use maximum
                    cpu = str(valid_cpu_values[-1])

        memory = DEFAULT_MEMORY
        memory_value_gib = None

        if resource_settings.memory is not None:
            memory_value_gib = resource_settings.get_memory(unit="GiB")

        final_memory_gib = self._validate_memory_for_cpu(cpu, memory_value_gib)

        if final_memory_gib is not None:
            if final_memory_gib == int(final_memory_gib):
                memory = f"{int(final_memory_gib)}Gi"
            else:
                memory = f"{final_memory_gib:.1f}Gi"

        return str(cpu), memory

    def _validate_memory_for_cpu(
        self, cpu: str, memory_gib: Optional[float]
    ) -> Optional[float]:
        """Validate and adjust memory allocation based on CPU requirements.

        GCP Cloud Run has minimum memory requirements per CPU configuration:
        - 1 CPU: 128 MiB minimum (0.125 GiB)
        - 2 CPU: 128 MiB minimum (0.125 GiB)
        - 4 CPU: 2 GiB minimum
        - 6 CPU: 4 GiB minimum
        - 8 CPU: 4 GiB minimum

        Args:
            cpu: CPU allocation as string (e.g., "1", "2", "4")
            memory_gib: Memory allocation in GiB (e.g., 2.0, 0.5, None)

        Returns:
            Adjusted memory allocation in GiB that meets minimum requirements, or None if no memory specified
        """
        if memory_gib is None:
            return None

        min_memory_per_cpu_gib = {
            1: 0.125,  # 128 MiB = 0.125 GiB
            2: 0.125,  # 128 MiB = 0.125 GiB
            4: 2.0,  # 2 GiB
            6: 4.0,  # 4 GiB
            8: 4.0,  # 4 GiB
        }

        # Handle fractional CPUs (< 1.0) - use minimum for 1 CPU
        cpu_float = float(cpu)
        if cpu_float < 1.0:
            cpu_int = 1
        else:
            cpu_int = int(cpu_float)

        required_memory_gib = min_memory_per_cpu_gib.get(cpu_int, 0.125)

        return max(memory_gib, required_memory_gib)

    def _convert_scaling_settings_to_gcp_format(
        self,
        resource_settings: ResourceSettings,
    ) -> Tuple[int, int]:
        """Convert ResourceSettings scaling to GCP Cloud Run format.

        Args:
            resource_settings: The resource settings from pipeline configuration.

        Returns:
            Tuple of (min_instances, max_instances) for GCP Cloud Run.
        """
        min_instances = DEFAULT_MIN_INSTANCES
        if resource_settings.min_replicas is not None:
            min_instances = resource_settings.min_replicas

        max_instances = DEFAULT_MAX_INSTANCES
        if resource_settings.max_replicas is not None:
            # ResourceSettings uses 0 to mean "no limit"
            # GCP Cloud Run needs a specific value, so we use the platform maximum
            if resource_settings.max_replicas == 0:
                max_instances = GCP_CLOUD_RUN_MAX_INSTANCES
            else:
                max_instances = resource_settings.max_replicas

        return min_instances, max_instances

    def _convert_concurrency_settings_to_gcp_format(
        self,
        resource_settings: ResourceSettings,
    ) -> int:
        """Convert ResourceSettings concurrency to GCP Cloud Run format.

        Args:
            resource_settings: The resource settings from pipeline configuration.

        Returns:
            The concurrency setting for GCP Cloud Run.
        """
        concurrency = DEFAULT_CONCURRENCY
        if resource_settings.max_concurrency is not None:
            concurrency = resource_settings.max_concurrency

        return concurrency

    def do_provision_deployment(
        self,
        deployment: DeploymentResponse,
        stack: "Stack",
        environment: Dict[str, str],
        secrets: Dict[str, str],
        timeout: int,
    ) -> DeploymentOperationalState:
        """Serve a pipeline as a Cloud Run service.

        Args:
            deployment: The deployment to serve.
            stack: The stack the pipeline will be deployed on.
            environment: Environment variables to set.
            secrets: Secret environment variables to set.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be provisioned.

        Returns:
            The operational state of the provisioned deployment.

        Raises:
            DeploymentProvisionError: If the deployment fails.
            DeployerError: If an unexpected error occurs.
        """
        snapshot = deployment.snapshot
        assert snapshot, "Pipeline snapshot not found"

        settings = cast(
            GCPDeployerSettings,
            self.get_settings(snapshot),
        )

        resource_settings = snapshot.pipeline_configuration.resource_settings

        cpu, memory = self._convert_resource_settings_to_gcp_format(
            resource_settings,
        )
        min_instances, max_instances = (
            self._convert_scaling_settings_to_gcp_format(
                resource_settings,
            )
        )
        concurrency = self._convert_concurrency_settings_to_gcp_format(
            resource_settings,
        )

        project_id = self.project_id

        service_name = self._get_service_name(
            deployment.name, deployment.id, settings.service_name_prefix
        )

        service_path = self._get_service_path(
            service_name, project_id, settings.location
        )

        # If a previous deployment of the same deployment exists but with
        # a different service name, location, or project, we need to clean up
        # the old service.
        existing_metadata = CloudRunDeploymentMetadata.from_deployment(
            deployment
        )

        if (
            existing_metadata.service_name
            and existing_metadata.location
            and existing_metadata.project_id
        ):
            existing_service_path = self._get_service_path(
                existing_metadata.service_name,
                existing_metadata.project_id,
                existing_metadata.location,
            )
            if existing_service_path != service_path:
                try:
                    self.do_deprovision_deployment(deployment, timeout)
                except DeploymentNotFoundError:
                    logger.warning(
                        f"Deployment '{deployment.name}' not found, "
                        f"skipping deprovision of existing Cloud Run service"
                    )
                except DeployerError as e:
                    logger.warning(
                        f"Failed to deprovision existing Cloud Run service for "
                        f"deployment '{deployment.name}': {e}"
                    )

        image = self.get_image(snapshot)

        entrypoint = DeploymentEntrypointConfiguration.get_entrypoint_command()
        arguments = DeploymentEntrypointConfiguration.get_entrypoint_arguments(
            **{
                DEPLOYMENT_ID_OPTION: deployment.id,
            }
        )

        env_vars, active_secrets = self._prepare_environment_variables(
            deployment, environment, secrets, settings, project_id
        )

        resources = run_v2.ResourceRequirements(
            limits={
                "cpu": cpu,
                "memory": memory,
            }
        )

        scaling = run_v2.RevisionScaling(
            min_instance_count=min_instances,
            max_instance_count=max_instances,
        )

        vpc_access = None
        if settings.vpc_connector:
            vpc_access = run_v2.VpcAccess(connector=settings.vpc_connector)

        container_port = (
            snapshot.pipeline_configuration.deployment_settings.uvicorn_port
        )
        container = run_v2.Container(
            image=image,
            command=entrypoint,
            args=arguments,
            env=env_vars,
            resources=resources,
            ports=[run_v2.ContainerPort(container_port=container_port)],
        )

        template = run_v2.RevisionTemplate(
            labels=settings.labels,
            annotations=settings.annotations,
            scaling=scaling,
            vpc_access=vpc_access,
            max_instance_request_concurrency=concurrency,
            timeout=f"{settings.timeout_seconds}s",
            service_account=settings.service_account,
            containers=[container],
            execution_environment=(
                run_v2.ExecutionEnvironment.EXECUTION_ENVIRONMENT_GEN2
                if settings.execution_environment == "gen2"
                else run_v2.ExecutionEnvironment.EXECUTION_ENVIRONMENT_GEN1
            ),
        )

        traffic = []
        for revision, percent in settings.traffic_allocation.items():
            if revision == "LATEST":
                traffic.append(
                    run_v2.TrafficTarget(
                        type_=run_v2.TrafficTargetAllocationType.TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST,
                        percent=percent,
                    )
                )
            else:
                traffic.append(
                    run_v2.TrafficTarget(
                        revision=revision,
                        percent=percent,
                    )
                )

        ingress_mapping = {
            "all": run_v2.IngressTraffic.INGRESS_TRAFFIC_ALL,
            "internal": run_v2.IngressTraffic.INGRESS_TRAFFIC_INTERNAL_ONLY,
            "internal-and-cloud-load-balancing": run_v2.IngressTraffic.INGRESS_TRAFFIC_INTERNAL_LOAD_BALANCER,
        }
        ingress = ingress_mapping.get(
            settings.ingress, run_v2.IngressTraffic.INGRESS_TRAFFIC_ALL
        )

        # Create the service (name should NOT be set for CreateServiceRequest)
        service = run_v2.Service(
            labels=self.get_labels(deployment, settings),
            annotations=settings.annotations,
            template=template,
            traffic=traffic,
            ingress=ingress,
            invoker_iam_disabled=settings.allow_unauthenticated,
        )

        try:
            existing_service = None
            try:
                existing_service = self.cloud_run_client.get_service(
                    name=service_path
                )
            except google_exceptions.NotFound:
                pass

            if existing_service:
                # Update existing service - need to set the name in the
                # CreateServiceRequest for updates
                service.name = service_path
                logger.debug(
                    f"Updating existing Cloud Run service for pipeline "
                    f"deployment '{deployment.name}'"
                )
                self.cloud_run_client.update_service(service=service)
            else:
                logger.debug(
                    f"Creating new Cloud Run service for deployment "
                    f"'{deployment.name}'"
                )
                parent = f"projects/{project_id}/locations/{settings.location}"
                # Create new service - name must not be set in the
                # CreateServiceRequest, using service_id instead
                self.cloud_run_client.create_service(
                    parent=parent, service=service, service_id=service_name
                )
                # Adding the name here for the operational state retrieval
                service.name = service_path

            return self._get_service_operational_state(
                service, project_id, settings.location, active_secrets
            )

        except google_exceptions.GoogleAPICallError as e:
            raise DeploymentProvisionError(
                f"Failed to deploy Cloud Run service for deployment "
                f"'{deployment.name}': {e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while provisioning deployment "
                f"'{deployment.name}': {e}"
            )

    def do_get_deployment_state(
        self,
        deployment: DeploymentResponse,
    ) -> DeploymentOperationalState:
        """Get information about a Cloud Run deployment.

        Args:
            deployment: The deployment to get information about.

        Returns:
            The operational state of the deployment.

        Raises:
            DeploymentNotFoundError: If the deployment is not found.
            RuntimeError: If the project ID or location is not found in the
                deployment metadata.
        """
        service = self._get_cloud_run_service(deployment)

        if service is None:
            raise DeploymentNotFoundError(
                f"Cloud Run service for deployment '{deployment.name}' "
                "not found"
            )

        existing_metadata = CloudRunDeploymentMetadata.from_deployment(
            deployment
        )

        if not existing_metadata.project_id or not existing_metadata.location:
            raise RuntimeError(
                f"Project ID or location not found in deployment metadata for "
                f"deployment '{deployment.name}'"
            )

        existing_secrets = self._get_secrets(deployment)

        return self._get_service_operational_state(
            service,
            existing_metadata.project_id,
            existing_metadata.location,
            existing_secrets,
        )

    def do_get_deployment_state_logs(
        self,
        deployment: DeploymentResponse,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Get the logs of a Cloud Run deployment.

        Args:
            deployment: The deployment to get the logs of.
            follow: If True, stream logs as they are written.
            tail: Only retrieve the last NUM lines of log output.

        Yields:
            The logs of the deployment.

        Raises:
            NotImplementedError: If log following is requested.
            DeploymentLogsNotFoundError: If the logs are not found.
            DeployerError: If an unexpected error occurs.
        """
        if follow:
            raise NotImplementedError(
                "Log following is not yet implemented for Cloud Run deployer"
            )

        try:
            existing_metadata = CloudRunDeploymentMetadata.from_deployment(
                deployment
            )
            service_name = existing_metadata.service_name
            if not service_name:
                assert deployment.snapshot, (
                    "Pipeline snapshot not set for deployment"
                )
                settings = cast(
                    GCPDeployerSettings,
                    self.get_settings(deployment.snapshot),
                )
                # We rely on the running service name, if a service is currently
                # active. If not, we fall back to the service name generated
                # from the current configuration.
                service_name = self._get_service_name(
                    deployment.name,
                    deployment.id,
                    settings.service_name_prefix,
                )

            filter_str = (
                'resource.type="cloud_run_revision" AND '
                f'resource.labels.service_name="{service_name}"'
            )

            entries = self.logging_client.list_entries(filter_=filter_str)

            log_lines = []
            for entry in entries:
                if hasattr(entry, "payload") and entry.payload:
                    timestamp = (
                        entry.timestamp.isoformat() if entry.timestamp else ""
                    )
                    log_line = f"[{timestamp}] {entry.payload}"
                    log_lines.append(log_line)

            if tail is not None and tail > 0:
                log_lines = log_lines[-tail:]

            for log_line in log_lines:
                yield log_line

        except google_exceptions.GoogleAPICallError as e:
            raise DeploymentLogsNotFoundError(
                f"Failed to retrieve logs for deployment "
                f"'{deployment.name}': {e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while retrieving logs for deployment "
                f"'{deployment.name}': {e}"
            )

    def do_deprovision_deployment(
        self,
        deployment: DeploymentResponse,
        timeout: int,
    ) -> Optional[DeploymentOperationalState]:
        """Deprovision a Cloud Run deployment.

        Args:
            deployment: The deployment to deprovision.
            timeout: The maximum time in seconds to wait for the pipeline
                deployment to be deprovisioned.

        Returns:
            The operational state of the deprovisioned deployment, or None if
            deletion is completed immediately.

        Raises:
            DeploymentNotFoundError: If the deployment is not found.
            DeploymentDeprovisionError: If the deprovision fails.
            DeployerError: If an unexpected error occurs.
            RuntimeError: If the service name, project ID or location is not
                found in the deployment metadata.
        """
        service = self._get_cloud_run_service(deployment)
        if service is None:
            raise DeploymentNotFoundError(
                f"Cloud Run service for deployment '{deployment.name}' not found"
            )

        try:
            existing_metadata = CloudRunDeploymentMetadata.from_deployment(
                deployment
            )
            if (
                not existing_metadata.service_name
                or not existing_metadata.project_id
                or not existing_metadata.location
            ):
                raise RuntimeError(
                    f"Service name, project ID or location not found in "
                    f"deployment metadata for deployment '{deployment.name}'"
                )

            service_path = self._get_service_path(
                existing_metadata.service_name,
                existing_metadata.project_id,
                existing_metadata.location,
            )

            logger.debug(
                f"Deleting Cloud Run service for deployment '{deployment.name}'"
            )

            self.cloud_run_client.delete_service(name=service_path)

            self._cleanup_deployment_secrets(deployment)

        except google_exceptions.NotFound:
            raise DeploymentNotFoundError(
                f"Cloud Run service for deployment '{deployment.name}' not found"
            )
        except google_exceptions.GoogleAPICallError as e:
            raise DeploymentDeprovisionError(
                f"Failed to delete Cloud Run service for deployment "
                f"'{deployment.name}': {e}"
            )
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while deleting deployment "
                f"'{deployment.name}': {e}"
            )

        return self.do_get_deployment_state(deployment)



# Original License:
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# New License:
#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the VertexAI orchestrator."""

import os
import re
import types
import urllib
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from uuid import UUID

from google.api_core import exceptions as google_exceptions
from google.cloud import aiplatform
from google.cloud.aiplatform.compat.services import (
    pipeline_service_client_v1beta1,
)
from google.cloud.aiplatform.compat.types import pipeline_job_v1beta1
from google.cloud.aiplatform_v1.types import JobState, PipelineState
from google.cloud.aiplatform_v1beta1.types.service_networking import (
    PscInterfaceConfig,
)
from google.protobuf import json_format
from google_cloud_pipeline_components.v1.custom_job.utils import (
    create_custom_training_job_from_component,
)
from kfp import dsl
from kfp.compiler import Compiler
from kfp.dsl.base_component import BaseComponent

from zenml import __version__
from zenml.config.resource_settings import ResourceSettings
from zenml.constants import (
    METADATA_ORCHESTRATOR_LOGS_URL,
    METADATA_ORCHESTRATOR_RUN_ID,
    METADATA_ORCHESTRATOR_URL,
    ORCHESTRATOR_DOCKER_IMAGE_KEY,
)
from zenml.entrypoints import StepEntrypointConfiguration
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.integrations.gcp import GCP_ARTIFACT_STORE_FLAVOR
from zenml.integrations.gcp.constants import (
    GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL,
    VERTEX_ENDPOINT_SUFFIX,
)
from zenml.integrations.gcp.flavors.vertex_orchestrator_flavor import (
    VertexOrchestratorConfig,
    VertexOrchestratorSettings,
)
from zenml.integrations.gcp.google_credentials_mixin import (
    GoogleCredentialsMixin,
)
from zenml.integrations.gcp.utils import (
    build_job_request,
    convert_job_state,
    monitor_job,
)
from zenml.integrations.gcp.vertex_custom_job_parameters import (
    VertexCustomJobParameters,
)
from zenml.io import fileio
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType, Uri
from zenml.orchestrators import ContainerizedOrchestrator, SubmissionResult
from zenml.orchestrators.publish_utils import publish_step_run_metadata
from zenml.orchestrators.utils import get_orchestrator_run_name
from zenml.pipelines.dynamic.entrypoint_configuration import (
    DynamicPipelineEntrypointConfiguration,
)
from zenml.stack.stack_validator import StackValidator
from zenml.step_operators.step_operator_entrypoint_configuration import (
    StepOperatorEntrypointConfiguration,
)
from zenml.utils.io_utils import get_global_config_directory

if TYPE_CHECKING:
    from zenml.config.base_settings import BaseSettings
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import (
        PipelineRunResponse,
        PipelineSnapshotResponse,
        ScheduleResponse,
        StepRunResponse,
    )
    from zenml.stack import Stack

logger = get_logger(__name__)
ENV_ZENML_VERTEX_RUN_ID = "ZENML_VERTEX_RUN_ID"
STEP_JOB_NAME_METADATA_KEY = "job_name"


def _clean_pipeline_name(pipeline_name: str) -> str:
    """Clean pipeline name to be a valid Vertex AI Pipeline name.

    Arguments:
        pipeline_name: pipeline name to be cleaned.

    Returns:
        Cleaned pipeline name.
    """
    pipeline_name = pipeline_name.lower()

    # This pattern matches anything that is not a lowercase letter,
    #  a number, or a dash
    pattern = r"[^a-z0-9-]"

    # Replace any characters matching the pattern with a dash
    return re.sub(pattern, "-", pipeline_name)


class VertexOrchestrator(ContainerizedOrchestrator, GoogleCredentialsMixin):
    """Orchestrator responsible for running pipelines on Vertex AI."""

    _pipeline_root: str
    _job_service_client: Optional[aiplatform.gapic.JobServiceClient] = None

    @property
    def config(self) -> VertexOrchestratorConfig:
        """Returns the `VertexOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(VertexOrchestratorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Vertex orchestrator.

        Returns:
            The settings class.
        """
        return VertexOrchestratorSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validates that the stack contains a container registry.

        Also validates that the artifact store is not local.

        Returns:
            A StackValidator instance.
        """

        def _validate_stack_requirements(stack: "Stack") -> Tuple[bool, str]:
            """Validates that all the stack components are not local.

            Args:
                stack: The stack to validate.

            Returns:
                A tuple of (is_valid, error_message).
            """
            # Validate that the container registry is not local.
            container_registry = stack.container_registry
            if container_registry and container_registry.config.is_local:
                return False, (
                    f"The Vertex orchestrator does not support local "
                    f"container registries. You should replace the component '"
                    f"{container_registry.name}' "
                    f"{container_registry.type.value} to a remote one."
                )

            # Validate that the rest of the components are not local.
            for stack_comp in stack.components.values():
                # For Forward compatibility a list of components is returned,
                # but only the first item is relevant for now
                # TODO: [server] make sure the ComponentModel actually has
                #  a local_path property or implement similar check
                local_path = stack_comp.local_path
                if not local_path:
                    continue
                return False, (
                    f"The '{stack_comp.name}' {stack_comp.type.value} is a "
                    f"local stack component. The Vertex AI Pipelines "
                    f"orchestrator requires that all the components in the "
                    f"stack used to execute the pipeline have to be not local, "
                    f"because there is no way for Vertex to connect to your "
                    f"local machine. You should use a flavor of "
                    f"{stack_comp.type.value} other than '"
                    f"{stack_comp.flavor}'."
                )

            # If the `pipeline_root` has not been defined in the orchestrator
            # configuration, and the artifact store is not a GCP artifact store,
            # then raise an error.
            if (
                not self.config.pipeline_root
                and stack.artifact_store.flavor != GCP_ARTIFACT_STORE_FLAVOR
            ):
                return False, (
                    f"The attribute `pipeline_root` has not been set and it "
                    f"cannot be generated using the path of the artifact store "
                    f"because it is not a "
                    f"`zenml.integrations.gcp.artifact_store.GCPArtifactStore`."
                    f" To solve this issue, set the `pipeline_root` attribute "
                    f"manually executing the following command: "
                    f"`zenml orchestrator update {stack.orchestrator.name} "
                    f'--pipeline_root="<Cloud Storage URI>"`.'
                )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_stack_requirements,
        )

    @property
    def root_directory(self) -> str:
        """Returns path to the root directory for files for this orchestrator.

        Returns:
            The path to the root directory for all files concerning this
            orchestrator.
        """
        return os.path.join(
            get_global_config_directory(), "vertex", str(self.id)
        )

    @property
    def pipeline_directory(self) -> str:
        """Returns path to directory where kubeflow pipelines files are stored.

        Returns:
            Path to the pipeline directory.
        """
        return os.path.join(self.root_directory, "pipelines")

    def get_job_service_client(self) -> aiplatform.gapic.JobServiceClient:
        """Get the job service client.

        Returns:
            The job service client.
        """
        if self.connector_has_expired():
            self._job_service_client = None

        if self._job_service_client is None:
            credentials, _ = self._get_authentication()
            client_options = {
                "api_endpoint": self.config.location + VERTEX_ENDPOINT_SUFFIX
            }
            self._job_service_client = aiplatform.gapic.JobServiceClient(
                credentials=credentials, client_options=client_options
            )
        return self._job_service_client

    def _create_container_component(
        self,
        image: str,
        command: List[str],
        arguments: List[str],
        component_name: str,
    ) -> BaseComponent:
        """Creates a container component for a Vertex pipeline.

        Args:
            image: The image to use for the component.
            command: The command to use for the component.
            arguments: The arguments to use for the component.
            component_name: The name of the component.

        Returns:
            The container component.
        """

        def dynamic_container_component() -> dsl.ContainerSpec:
            """Dynamic container component.

            Returns:
                The dynamic container component.
            """
            return dsl.ContainerSpec(
                image=image,
                command=command,
                args=arguments,
            )

        new_container_spec_func = types.FunctionType(
            dynamic_container_component.__code__,
            dynamic_container_component.__globals__,
            name=component_name,
            argdefs=dynamic_container_component.__defaults__,
            closure=dynamic_container_component.__closure__,
        )
        pipeline_task = dsl.container_component(new_container_spec_func)
        return pipeline_task

    def _convert_to_custom_training_job(
        self,
        component: BaseComponent,
        settings: VertexOrchestratorSettings,
        environment: Dict[str, str],
    ) -> BaseComponent:
        """Convert a component to a custom training job component.

        Args:
            component: The component to convert.
            settings: The settings for the custom training job.
            environment: The environment variables to set in the custom
                training job.

        Returns:
            The custom training job component.
        """
        custom_job_parameters = (
            settings.custom_job_parameters or VertexCustomJobParameters()
        )
        if (
            custom_job_parameters.persistent_resource_id
            and not custom_job_parameters.service_account
        ):
            # Persistent resources require an explicit service account, but
            # none was provided in the custom job parameters. We try to fall
            # back to the workload service account.
            custom_job_parameters.service_account = (
                self.config.workload_service_account
            )

        # Create a dictionary of explicit parameters
        params = custom_job_parameters.model_dump(
            exclude_none=True, exclude={"additional_training_job_args"}
        )

        # Remove None values to let defaults be set by the function
        params = {k: v for k, v in params.items() if v is not None}

        # Add environment variables
        params["env"] = [
            {"name": key, "value": value} for key, value in environment.items()
        ]

        # Check if any advanced parameters will override explicit parameters
        if custom_job_parameters.additional_training_job_args:
            overridden_params = set(params.keys()) & set(
                custom_job_parameters.additional_training_job_args.keys()
            )
            if overridden_params:
                logger.warning(
                    f"The following explicit parameters are being overridden by values in "
                    f"additional_training_job_args: {', '.join(overridden_params)}. "
                    f"This may lead to unexpected behavior. Consider using either explicit "
                    f"parameters or additional_training_job_args, but not both for the same parameters."
                )

        # Add any advanced parameters - these will override explicit parameters if provided
        params.update(custom_job_parameters.additional_training_job_args)

        # Add other parameters from orchestrator config if not already in params
        if self.config.network and "network" not in params:
            params["network"] = self.config.network

        if (
            self.config.encryption_spec_key_name
            and "encryption_spec_key_name" not in params
        ):
            params["encryption_spec_key_name"] = (
                self.config.encryption_spec_key_name
            )
        if (
            self.config.workload_service_account
            and "service_account" not in params
        ):
            params["service_account"] = self.config.workload_service_account

        custom_job_component = create_custom_training_job_from_component(
            component_spec=component,
            **params,
        )

        return custom_job_component

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Raises:
            ValueError: If a schedule without a cron expression is passed.

        Returns:
            Optional submission result.
        """
        if snapshot.schedule:
            if snapshot.schedule.catchup or snapshot.schedule.interval_second:
                logger.warning(
                    "Vertex orchestrator only uses schedules with the "
                    "`cron_expression` property, with optional `start_time` "
                    "and/or `end_time`. All other properties are ignored."
                )
            if snapshot.schedule.cron_expression is None:
                raise ValueError(
                    "Property `cron_expression` must be set when passing "
                    "schedule to a Vertex orchestrator."
                )

        orchestrator_run_name = get_orchestrator_run_name(
            pipeline_name=snapshot.pipeline_configuration.name
        )
        # If the `pipeline_root` has not been defined in the orchestrator
        # configuration,
        # try to create it from the artifact store if it is a
        # `GCPArtifactStore`.
        if not self.config.pipeline_root:
            artifact_store = stack.artifact_store
            self._pipeline_root = f"{artifact_store.path.rstrip('/')}/vertex_pipeline_root/{snapshot.pipeline_configuration.name}/{orchestrator_run_name}"
            logger.info(
                "The attribute `pipeline_root` has not been set in the "
                "orchestrator configuration. One has been generated "
                "automatically based on the path of the `GCPArtifactStore` "
                "artifact store in the stack used to execute the pipeline. "
                "The generated `pipeline_root` is `%s`.",
                self._pipeline_root,
            )
        else:
            self._pipeline_root = self.config.pipeline_root

        def _create_dynamic_pipeline() -> Any:
            """Create a dynamic pipeline including each step.

            Returns:
                pipeline_func
            """
            step_name_to_dynamic_component: Dict[str, BaseComponent] = {}

            for step_name, step in snapshot.step_configurations.items():
                image = self.get_image(
                    snapshot=snapshot,
                    step_name=step_name,
                )
                command = StepEntrypointConfiguration.get_entrypoint_command()
                arguments = (
                    StepEntrypointConfiguration.get_entrypoint_arguments(
                        step_name=step_name,
                        snapshot_id=snapshot.id,
                    )
                )
                component = self._create_container_component(
                    image, command, arguments, step_name
                )
                step_settings = cast(
                    VertexOrchestratorSettings, self.get_settings(step)
                )
                pod_settings = step_settings.pod_settings
                if pod_settings:
                    ignored_fields = pod_settings.model_fields_set - {
                        "node_selectors"
                    }
                    if ignored_fields:
                        logger.warning(
                            f"The following pod settings are not supported in "
                            f"Vertex with Vertex Pipelines 2.x and will be "
                            f"ignored: {list(ignored_fields)}."
                        )

                    for key in pod_settings.node_selectors:
                        if (
                            key
                            != GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                        ):
                            logger.warning(
                                "Vertex only allows the %s node selector, "
                                "ignoring the node selector %s.",
                                GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL,
                                key,
                            )

                step_name_to_dynamic_component[step_name] = component

            @dsl.pipeline(  # type: ignore[misc]
                display_name=orchestrator_run_name,
            )
            def dynamic_pipeline() -> None:
                """Dynamic pipeline."""
                # iterate through the components one by one
                # (from step_name_to_dynamic_component)
                for (
                    component_name,
                    component,
                ) in step_name_to_dynamic_component.items():
                    # for each component, check to see what other steps are
                    # upstream of it
                    step = snapshot.step_configurations[component_name]
                    upstream_step_components = [
                        step_name_to_dynamic_component[upstream_step_name]
                        for upstream_step_name in step.spec.upstream_steps
                    ]

                    step_settings = cast(
                        VertexOrchestratorSettings, self.get_settings(step)
                    )

                    use_custom_training_job = (
                        step_settings.custom_job_parameters is not None
                    )

                    step_environment = step_environments[component_name]
                    step_environment[ENV_ZENML_VERTEX_RUN_ID] = (
                        dsl.PIPELINE_JOB_NAME_PLACEHOLDER
                    )

                    if use_custom_training_job:
                        if not step.config.resource_settings.empty:
                            logger.warning(
                                "Ignoring resource settings because "
                                "the step is running as a custom training job. "
                                "Use `custom_job_parameters.machine_type` "
                                "to configure the machine type instead."
                            )
                        if step_settings.node_selector_constraint:
                            logger.warning(
                                "Ignoring node selector constraint because "
                                "the step is running as a custom training job. "
                                "Use `custom_job_parameters.accelerator_type` "
                                "to configure the accelerator type instead."
                            )
                        component = self._convert_to_custom_training_job(
                            component,
                            settings=step_settings,
                            environment=step_environment,
                        )
                        task = (
                            component()
                            .set_display_name(name=component_name)
                            .set_caching_options(enable_caching=False)
                            .after(*upstream_step_components)
                        )
                    else:
                        task = (
                            component()
                            .set_display_name(
                                name=component_name,
                            )
                            .set_caching_options(enable_caching=False)
                            .after(*upstream_step_components)
                        )
                        for key, value in step_environment.items():
                            task = task.set_env_variable(name=key, value=value)

                        pod_settings = step_settings.pod_settings

                        node_selector_constraint: Optional[Tuple[str, str]] = (
                            None
                        )
                        if pod_settings and (
                            GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                            in pod_settings.node_selectors.keys()
                        ):
                            node_selector_constraint = (
                                GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL,
                                pod_settings.node_selectors[
                                    GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL
                                ],
                            )
                        elif step_settings.node_selector_constraint:
                            node_selector_constraint = (
                                GKE_ACCELERATOR_NODE_SELECTOR_CONSTRAINT_LABEL,
                                step_settings.node_selector_constraint[1],
                            )

                        self._configure_container_resources(
                            dynamic_component=task,
                            resource_settings=step.config.resource_settings,
                            node_selector_constraint=node_selector_constraint,
                        )

            return dynamic_pipeline

        # Save the generated pipeline to a file.
        fileio.makedirs(self.pipeline_directory)
        pipeline_file_path = os.path.join(
            self.pipeline_directory,
            f"{orchestrator_run_name}.json",
        )

        # Compile the pipeline using the Kubeflow SDK V2 compiler that allows
        # to generate a JSON representation of the pipeline that can be later
        # upload to Vertex AI Pipelines service.
        Compiler().compile(
            pipeline_func=_create_dynamic_pipeline(),
            package_path=pipeline_file_path,
            pipeline_name=_clean_pipeline_name(
                snapshot.pipeline_configuration.name
            ),
        )

        logger.info(
            "Writing Vertex workflow definition to `%s`.", pipeline_file_path
        )

        settings = cast(
            VertexOrchestratorSettings, self.get_settings(snapshot)
        )

        return self._upload_and_run_pipeline(
            pipeline_name=snapshot.pipeline_configuration.name,
            pipeline_file_path=pipeline_file_path,
            run_name=orchestrator_run_name,
            settings=settings,
            schedule=snapshot.schedule,
        )

    def submit_dynamic_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a dynamic pipeline to the orchestrator.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment.
            placeholder_run: An optional placeholder run.

        Raises:
            RuntimeError: If the snapshot contains a schedule.

        Returns:
            Optional submission result.
        """
        if snapshot.schedule:
            raise RuntimeError(
                "Scheduling dynamic pipelines is not supported for the "
                "Vertex orchestrator yet."
            )

        settings = cast(
            VertexOrchestratorSettings, self.get_settings(snapshot)
        )

        command = (
            DynamicPipelineEntrypointConfiguration.get_entrypoint_command()
        )
        args = DynamicPipelineEntrypointConfiguration.get_entrypoint_arguments(
            snapshot_id=snapshot.id,
            run_id=placeholder_run.id if placeholder_run else None,
        )

        image = self.get_image(snapshot=snapshot)
        labels = settings.labels.copy()
        labels["source"] = f"zenml-{__version__.replace('.', '_')}"

        job_request = build_job_request(
            display_name=get_orchestrator_run_name(
                pipeline_name=snapshot.pipeline_configuration.name
            ),
            image=image,
            entrypoint_command=command + args,
            custom_job_settings=settings.custom_job_parameters
            or VertexCustomJobParameters(),
            resource_settings=snapshot.pipeline_configuration.resource_settings,
            environment=environment,
            labels=labels,
            encryption_spec_key_name=self.config.encryption_spec_key_name,
            service_account=self.config.workload_service_account,
            network=self.config.network,
        )

        client = self.get_job_service_client()
        parent = (
            f"projects/{self.gcp_project_id}/locations/{self.config.location}"
        )
        job_model = client.create_custom_job(
            parent=parent, custom_job=job_request
        )

        _wait_for_completion = None
        if settings.synchronous:

            def _wait_for_completion() -> None:
                logger.info("Waiting for the VertexAI job to finish...")
                monitor_job(
                    job_id=job_model.name,
                    get_client=self.get_job_service_client,
                )
                logger.info("VertexAI job completed successfully.")

        credentials, project_id = self._get_authentication()
        job = aiplatform.CustomJob.get(
            job_model.name,
            project=project_id,
            location=self.config.location,
            credentials=credentials,
        )
        metadata = self.compute_metadata(job)

        logger.info("View the Vertex job at %s", job._dashboard_uri())

        return SubmissionResult(
            wait_for_completion=_wait_for_completion,
            metadata=metadata,
        )

    def submit_isolated_step(
        self, step_run_info: "StepRunInfo", environment: Dict[str, str]
    ) -> None:
        """Runs an isolated step on Vertex.

        Args:
            step_run_info: The step run information.
            environment: The environment variables to set.
        """
        settings = cast(
            VertexOrchestratorSettings, self.get_settings(step_run_info)
        )

        image = step_run_info.get_image(key=ORCHESTRATOR_DOCKER_IMAGE_KEY)
        command = StepOperatorEntrypointConfiguration.get_entrypoint_command()
        args = StepOperatorEntrypointConfiguration.get_entrypoint_arguments(
            step_name=step_run_info.pipeline_step_name,
            snapshot_id=(step_run_info.snapshot.id),
            step_run_id=str(step_run_info.step_run_id),
        )

        labels = settings.labels.copy()
        labels["source"] = f"zenml-{__version__.replace('.', '_')}"
        labels["step-run-id"] = str(step_run_info.step_run_id)

        job_request = build_job_request(
            display_name=f"{step_run_info.run_name}-{step_run_info.pipeline_step_name}",
            image=image,
            entrypoint_command=command + args,
            custom_job_settings=settings.custom_job_parameters
            or VertexCustomJobParameters(),
            resource_settings=step_run_info.config.resource_settings,
            environment=environment,
            labels=labels,
            encryption_spec_key_name=self.config.encryption_spec_key_name,
            service_account=self.config.workload_service_account,
            network=self.config.network,
        )

        client = self.get_job_service_client()
        parent = (
            f"projects/{self.gcp_project_id}/locations/{self.config.location}"
        )
        logger.info(
            "Submitting custom job='%s', path='%s' to Vertex AI Training.",
            job_request["display_name"],
            parent,
        )
        job = client.create_custom_job(parent=parent, custom_job=job_request)
        publish_step_run_metadata(
            step_run_info.step_run_id,
            {
                self.id: {STEP_JOB_NAME_METADATA_KEY: job.name},
            },
        )
        # Include the metadata in the step run response to avoid an extra
        # API call
        step_run_info.step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY] = (
            job.name
        )

    def get_isolated_step_status(
        self, step_run: "StepRunResponse"
    ) -> ExecutionStatus:
        """Gets the status of an isolated step on Vertex.

        Args:
            step_run: The step run.

        Returns:
            The status of the step run.
        """
        client = self.get_job_service_client()
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        job = client.get_custom_job(name=job_name)
        return convert_job_state(job.state)

    def stop_isolated_step(self, step_run: "StepRunResponse") -> None:
        """Stops an isolated step on Vertex.

        Args:
            step_run: The step run.
        """
        client = self.get_job_service_client()
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        client.cancel_custom_job(name=job_name)

    def _upload_and_run_pipeline(
        self,
        pipeline_name: str,
        pipeline_file_path: str,
        run_name: str,
        settings: VertexOrchestratorSettings,
        schedule: Optional["ScheduleResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Uploads and run the pipeline on the Vertex AI Pipelines service.

        Args:
            pipeline_name: Name of the pipeline.
            pipeline_file_path: Path of the JSON file containing the compiled
                Kubeflow pipeline (compiled with Kubeflow SDK v2).
            run_name: Orchestrator run name.
            settings: Pipeline level settings for this orchestrator.
            schedule: The schedule the pipeline will run on.

        Raises:
            RuntimeError: If the Vertex Orchestrator fails to provision or any
                other Runtime errors.

        Returns:
            Optional submission result.
        """
        # We have to replace the hyphens in the run name with underscores
        # and lower case the string, because the Vertex AI Pipelines service
        # requires this format.
        job_id = _clean_pipeline_name(run_name)

        # Get the credentials that would be used to create the Vertex AI
        # Pipelines job.
        credentials, project_id = self._get_authentication()

        # Instantiate the Vertex AI Pipelines job
        run = aiplatform.PipelineJob(
            display_name=pipeline_name,
            template_path=pipeline_file_path,
            job_id=job_id,
            pipeline_root=self._pipeline_root,
            parameter_values=None,
            enable_caching=False,
            encryption_spec_key_name=self.config.encryption_spec_key_name,
            labels=settings.labels,
            credentials=credentials,
            project=project_id,
            location=self.config.location,
        )

        if self.config.workload_service_account:
            logger.info(
                "The Vertex AI Pipelines job workload will be executed "
                "using the `%s` "
                "service account.",
                self.config.workload_service_account,
            )
        if self.config.network:
            logger.info(
                "The Vertex AI Pipelines job will be peered with the `%s` "
                "network.",
                self.config.network,
            )

        if self.config.private_service_connect:
            # The PSC setting isn't yet part of the stable v1 API. We need to
            # temporarily hack the aiplatform.PipelineJob object in two places:
            # * to use the v1beta1 PipelineJob primitive which supports the
            #   psc_interface_config field instead of the v1 PipelineJob
            #   primitive.
            # * to use the v1beta1 PipelineServiceClient instead of the v1
            #   PipelineServiceClient.
            #
            # We achieve the first by converting the v1 PipelineJob to a
            # v1beta1 PipelineJob and the second by replacing the v1
            # PipelineServiceClient with a v1beta1 PipelineServiceClient.
            #
            # TODO: Remove this once the v1 stable API is updated to support
            # the PSC setting.
            pipeline_job_dict = json_format.MessageToDict(
                run._gca_resource._pb, preserving_proto_field_name=True
            )
            run._gca_resource = pipeline_job_v1beta1.PipelineJob(
                **pipeline_job_dict,
                psc_interface_config=PscInterfaceConfig(
                    network_attachment=self.config.private_service_connect,
                ),
            )
            run.api_client = (
                pipeline_service_client_v1beta1.PipelineServiceClient(
                    credentials=run.credentials,
                    client_options={
                        "api_endpoint": run.api_client.api_endpoint,
                    },
                )
            )

        try:
            if schedule:
                logger.info(
                    "Scheduling job using native Vertex AI Pipelines "
                    "scheduling..."
                )
                run.create_schedule(
                    display_name=schedule.name,
                    cron=schedule.cron_expression,
                    start_time=schedule.utc_start_time,
                    end_time=schedule.utc_end_time,
                    service_account=self.config.workload_service_account,
                    network=self.config.network,
                )
                return None

            else:
                logger.info(
                    "No schedule detected. Creating one-off Vertex job..."
                )
                logger.info(
                    "Submitting pipeline job with job_id `%s` to Vertex AI "
                    "Pipelines service.",
                    job_id,
                )

                # Submit the job to Vertex AI Pipelines service.
                run.submit(
                    service_account=self.config.workload_service_account,
                    network=self.config.network,
                )
                logger.info(
                    "View the Vertex AI Pipelines job at %s",
                    run._dashboard_uri(),
                )

                _wait_for_completion = None

                if settings.synchronous:

                    def _wait_for_completion() -> None:
                        logger.info(
                            "Waiting for the Vertex AI Pipelines job to finish..."
                        )
                        run.wait()
                        logger.info(
                            "Vertex AI Pipelines job completed successfully."
                        )

                return SubmissionResult(
                    metadata=self.compute_metadata(run),
                    wait_for_completion=_wait_for_completion,
                )

        except google_exceptions.ClientError as e:
            logger.error("Failed to create the Vertex AI Pipelines job: %s", e)
            raise RuntimeError(
                f"Failed to create the Vertex AI Pipelines job: {e}"
            )
        except RuntimeError as e:
            logger.error(
                "The Vertex AI Pipelines job execution has failed: %s", e
            )
            raise

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Raises:
            RuntimeError: If the orchestrator run id cannot be read from the
                environment.

        Returns:
            The orchestrator run id.
        """
        for env in [ENV_ZENML_VERTEX_RUN_ID, "CLOUD_ML_JOB_ID"]:
            if env in os.environ:
                return os.environ[env]

        raise RuntimeError(
            "Unable to get orchestrator run id from environment."
        )

    def get_pipeline_run_metadata(
        self, run_id: UUID
    ) -> Dict[str, "MetadataType"]:
        """Get general component-specific metadata for a pipeline run.

        Args:
            run_id: The ID of the pipeline run.

        Returns:
            A dictionary of metadata.
        """
        if ENV_ZENML_VERTEX_RUN_ID in os.environ:
            # Static pipeline -> Pipeline job
            run_url = (
                f"https://console.cloud.google.com/vertex-ai/locations/"
                f"{self.config.location}/pipelines/runs/"
                f"{self.get_orchestrator_run_id()}"
            )
        else:
            # Dynamic pipeline -> Custom job
            run_url = (
                f"https://console.cloud.google.com/vertex-ai/locations/"
                f"{self.config.location}/training/"
                f"{self.get_orchestrator_run_id()}"
            )

        if self.config.project:
            run_url += f"?project={self.config.project}"

        return {
            METADATA_ORCHESTRATOR_RUN_ID: self.get_orchestrator_run_id(),
            METADATA_ORCHESTRATOR_URL: Uri(run_url),
        }

    def _configure_container_resources(
        self,
        dynamic_component: dsl.PipelineTask,
        resource_settings: "ResourceSettings",
        node_selector_constraint: Optional[Tuple[str, str]] = None,
    ) -> dsl.PipelineTask:
        """Adds resource requirements to the container.

        Args:
            dynamic_component: The dynamic component to add the resource
                settings to.
            resource_settings: The resource settings to use for this
                container.
            node_selector_constraint: Node selector constraint to apply to
                the container.

        Returns:
            The dynamic component with the resource settings applied.
        """
        # Set optional CPU, RAM and GPU constraints for the pipeline
        cpu_limit = None
        if resource_settings:
            cpu_limit = resource_settings.cpu_count or self.config.cpu_limit

        if cpu_limit is not None:
            dynamic_component = dynamic_component.set_cpu_limit(str(cpu_limit))

        memory_limit = (
            resource_settings.memory[:-1]
            if resource_settings.memory
            else self.config.memory_limit
        )
        if memory_limit is not None:
            dynamic_component = dynamic_component.set_memory_limit(
                memory_limit
            )

        gpu_limit = (
            resource_settings.gpu_count
            if resource_settings.gpu_count is not None
            else self.config.gpu_limit
        )

        if node_selector_constraint:
            _, value = node_selector_constraint
            if gpu_limit is not None and gpu_limit > 0:
                dynamic_component = (
                    dynamic_component.set_accelerator_type(value)
                    .set_accelerator_limit(gpu_limit)
                    .set_gpu_limit(gpu_limit)
                )
            else:
                logger.warning(
                    "Accelerator type %s specified, but the GPU limit is not "
                    "set or set to 0. The accelerator type will be ignored. "
                    "To fix this warning, either remove the specified "
                    "accelerator type or set the `gpu_count` using the "
                    "ResourceSettings (https://docs.zenml.io/user-guides/tutorial/distributed-training)."
                )

        return dynamic_component

    def fetch_status(
        self, run: "PipelineRunResponse", include_steps: bool = False
    ) -> Tuple[
        Optional[ExecutionStatus], Optional[Dict[str, ExecutionStatus]]
    ]:
        """Refreshes the status of a specific pipeline run.

        Args:
            run: The run that was executed by this orchestrator.
            include_steps: Whether to fetch steps.

        Returns:
            A tuple of (pipeline_status, step_statuses_dict).
            Step statuses are not supported for Vertex, so step_statuses_dict will always be None.

        Raises:
            AssertionError: If the run was not executed by to this orchestrator.
            ValueError: If it fetches an unknown state or if we can not fetch
                the orchestrator run ID.
        """
        # Make sure that the stack exists and is accessible
        if run.stack is None:
            raise ValueError(
                "The stack that the run was executed on is not available "
                "anymore."
            )

        # Make sure that the run belongs to this orchestrator
        assert (
            self.id
            == run.stack.components[StackComponentType.ORCHESTRATOR][0].id
        )

        # Fetch the status of the PipelineJob
        if METADATA_ORCHESTRATOR_RUN_ID in run.run_metadata:
            run_id = run.run_metadata[METADATA_ORCHESTRATOR_RUN_ID]
        elif run.orchestrator_run_id is not None:
            run_id = run.orchestrator_run_id
        else:
            raise ValueError(
                "Can not find the orchestrator run ID, thus can not fetch "
                "the status."
            )

        credentials, project_id = self._get_authentication()
        if run.snapshot and run.snapshot.is_dynamic:
            status = aiplatform.CustomJob.get(
                run_id,
                project=project_id,
                location=self.config.location,
                credentials=credentials,
            ).state

            if status in [
                JobState.JOB_STATE_QUEUED,
                JobState.JOB_STATE_PENDING,
            ]:
                pipeline_status = ExecutionStatus.PROVISIONING
            elif status in [
                JobState.JOB_STATE_RUNNING,
                JobState.JOB_STATE_PAUSED,
                JobState.JOB_STATE_UPDATING,
            ]:
                pipeline_status = ExecutionStatus.RUNNING
            elif status == JobState.JOB_STATE_SUCCEEDED:
                pipeline_status = ExecutionStatus.COMPLETED
            elif status == JobState.JOB_STATE_CANCELLING:
                pipeline_status = ExecutionStatus.STOPPING
            elif status == JobState.JOB_STATE_CANCELLED:
                pipeline_status = ExecutionStatus.STOPPED
            elif status in [
                JobState.JOB_STATE_FAILED,
                JobState.JOB_STATE_EXPIRED,
            ]:
                pipeline_status = ExecutionStatus.FAILED
            else:
                pipeline_status = run.status
        else:
            status = aiplatform.PipelineJob.get(
                run_id,
                project=project_id,
                location=self.config.location,
                credentials=credentials,
            ).state

            # Map the potential outputs to ZenML ExecutionStatus. Potential values:
            # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sagemaker/client/describe_pipeline_execution.html#
            if status == PipelineState.PIPELINE_STATE_UNSPECIFIED:
                pipeline_status = run.status
            elif status in [
                PipelineState.PIPELINE_STATE_QUEUED,
                PipelineState.PIPELINE_STATE_PENDING,
            ]:
                pipeline_status = ExecutionStatus.PROVISIONING
            elif status in [
                PipelineState.PIPELINE_STATE_RUNNING,
                PipelineState.PIPELINE_STATE_PAUSED,
            ]:
                pipeline_status = ExecutionStatus.RUNNING
            elif status == PipelineState.PIPELINE_STATE_SUCCEEDED:
                pipeline_status = ExecutionStatus.COMPLETED
            elif status == PipelineState.PIPELINE_STATE_CANCELLING:
                pipeline_status = ExecutionStatus.STOPPING
            elif status == PipelineState.PIPELINE_STATE_CANCELLED:
                pipeline_status = ExecutionStatus.STOPPED
            elif status == PipelineState.PIPELINE_STATE_FAILED:
                pipeline_status = ExecutionStatus.FAILED
            else:
                raise ValueError("Unknown status for the pipeline job.")

        # Vertex doesn't support step-level status fetching yet
        return pipeline_status, None

    def compute_metadata(
        self, job: Union[aiplatform.PipelineJob, aiplatform.CustomJob]
    ) -> Dict[str, MetadataType]:
        """Generate run metadata based on the Vertex job.

        Args:
            job: The job.

        Returns:
            A dictionary of metadata related to the pipeline run.
        """
        metadata: Dict[str, MetadataType] = {}

        if run_id := self._compute_orchestrator_run_id(job):
            metadata[METADATA_ORCHESTRATOR_RUN_ID] = run_id

        if orchestrator_url := self._compute_orchestrator_url(job):
            metadata[METADATA_ORCHESTRATOR_URL] = Uri(orchestrator_url)

        if logs_url := self._compute_orchestrator_logs_url(job):
            metadata[METADATA_ORCHESTRATOR_LOGS_URL] = Uri(logs_url)

        return metadata

    @staticmethod
    def _compute_orchestrator_url(
        job: Union[aiplatform.PipelineJob, aiplatform.CustomJob],
    ) -> Optional[str]:
        """Generate the Orchestrator Dashboard URL.

        Args:
            job: The job.

        Returns:
            The Vertex Dashboard URL for the job.
        """
        try:
            if uri := job._dashboard_uri():
                return Uri(uri)
        except Exception as e:
            logger.warning(
                "There was an issue while extracting the job dashboard URL: %s",
                e,
            )
        return None

    @staticmethod
    def _compute_orchestrator_logs_url(
        job: Union[aiplatform.PipelineJob, aiplatform.CustomJob],
    ) -> Optional[str]:
        """Generate the Logs Explorer URL.

        Args:
            job: The job.

        Returns:
            The Logs Explorer URL for the job.
        """
        try:
            if isinstance(job, aiplatform.PipelineJob):
                query = f"""
                resource.type="aiplatform.googleapis.com/PipelineJob"
                resource.labels.pipeline_job_id="{job.job_id}"
                """
            else:
                query = f'resource.labels.job_id="{job.name}"'

            query = urllib.parse.quote(query)
        except Exception as e:
            logger.warning(
                f"There was an issue while extracting the logs url: {e}"
            )
            return None
        else:
            return f"https://console.cloud.google.com/logs/query?project={job.project}&query={query}"

    @staticmethod
    def _compute_orchestrator_run_id(
        job: Union[aiplatform.PipelineJob, aiplatform.CustomJob],
    ) -> Optional[str]:
        """Fetch the orchestrator run ID.

        Args:
            job: The job.

        Returns:
            The orchestrator run ID.
        """
        try:
            if isinstance(job, aiplatform.PipelineJob):
                return str(job.job_id)
            else:
                return str(job.name)
        except Exception as e:
            logger.warning(
                "There was an issue while extracting the orchestrator run ID: %s",
                e,
            )
            return None



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Datadog log store implementation."""

from datetime import datetime
from typing import Any, Dict, List, Optional, cast

import requests

from zenml.enums import LoggingLevels
from zenml.log_stores.datadog.datadog_flavor import DatadogLogStoreConfig
from zenml.log_stores.datadog.datadog_log_exporter import DatadogLogExporter
from zenml.log_stores.otel.otel_log_store import OtelLogStore
from zenml.logger import get_logger
from zenml.models import LogsResponse
from zenml.utils.logging_utils import LogEntry

logger = get_logger(__name__)


class DatadogLogStore(OtelLogStore):
    """Log store that exports logs to Datadog.

    This implementation extends OtelLogStore and configures it to send logs
    to Datadog's HTTP intake API.
    """

    _datadog_exporter: Optional[DatadogLogExporter] = None

    @property
    def config(self) -> DatadogLogStoreConfig:
        """Returns the configuration of the Datadog log store.

        Returns:
            The configuration.
        """
        return cast(DatadogLogStoreConfig, self._config)

    def get_exporter(self) -> DatadogLogExporter:
        """Get the Datadog log exporter.

        Returns:
            DatadogExporter with the proper configuration.
        """
        if not self._datadog_exporter:
            headers = {
                "dd-api-key": self.config.api_key.get_secret_value(),
                "dd-application-key": self.config.application_key.get_secret_value(),
            }
            if self.config.headers:
                headers.update(self.config.headers)

            self._datadog_exporter = DatadogLogExporter(
                endpoint=self.config.endpoint,
                headers=headers,
                certificate_file=self.config.certificate_file,
                client_key_file=self.config.client_key_file,
                client_certificate_file=self.config.client_certificate_file,
                compression=self.config.compression,
            )
        return self._datadog_exporter

    def fetch(
        self,
        logs_model: "LogsResponse",
        limit: int,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ) -> List["LogEntry"]:
        """Fetch logs from Datadog's API.

        This method queries Datadog's Logs API to retrieve logs for the
        specified pipeline run and step. It automatically paginates through
        results to fetch up to the requested limit.

        Args:
            logs_model: The logs model containing run and step metadata.
            start_time: Filter logs after this time.
            end_time: Filter logs before this time.
            limit: Maximum number of log entries to return.

        Returns:
            List of log entries from Datadog.
        """
        query_parts = [
            f"service:{self.config.service_name}",
            f"@zenml.log.id:{logs_model.id}",
        ]

        query = " ".join(query_parts)

        api_endpoint = (
            f"https://api.{self.config.site}/api/v2/logs/events/search"
        )
        headers = {
            "DD-API-KEY": self.config.api_key.get_secret_value(),
            "DD-APPLICATION-KEY": self.config.application_key.get_secret_value(),
            "Content-Type": "application/json",
        }

        log_entries: List[LogEntry] = []
        cursor: Optional[str] = None
        remaining = limit

        try:
            while remaining > 0:
                # Datadog API limit is 1000 per request
                page_limit = min(remaining, 1000)

                body: Dict[str, Any] = {
                    "filter": {
                        "query": query,
                        "from": (
                            start_time.isoformat()
                            if start_time
                            else logs_model.created.isoformat()
                        ),
                        "to": (
                            end_time.isoformat()
                            if end_time
                            else datetime.now().astimezone().isoformat()
                        ),
                    },
                    "page": {
                        "limit": page_limit,
                    },
                    "sort": "timestamp",
                }

                if cursor:
                    body["page"]["cursor"] = cursor

                response = requests.post(
                    api_endpoint,
                    headers=headers,
                    json=body,
                    timeout=30,
                )

                if response.status_code != 200:
                    logger.error(
                        f"Failed to fetch logs from Datadog: "
                        f"{response.status_code} - {response.text[:200]}"
                    )
                    break

                data = response.json()
                logs = data.get("data", [])

                if not logs:
                    break

                for log in logs:
                    entry = self._parse_log_entry(log)
                    if entry:
                        log_entries.append(entry)

                remaining -= len(logs)

                # Get cursor for next page
                cursor = data.get("meta", {}).get("page", {}).get("after")
                if not cursor:
                    break

            logger.debug(f"Fetched {len(log_entries)} logs from Datadog")
            return log_entries

        except Exception as e:
            logger.exception(f"Error fetching logs from Datadog: {e}")
            return log_entries  # Return what we have so far

    def _parse_log_entry(self, log: Dict[str, Any]) -> Optional[LogEntry]:
        """Parse a single log entry from Datadog's API response.

        Args:
            log: The log data from Datadog's API.

        Returns:
            A LogEntry object, or None if parsing fails.
        """
        try:
            log_fields = log.get("attributes", {})
            message = log_fields.get("message", "")
            nested_attrs = log_fields.get("attributes", {})

            if exc_info := nested_attrs.get("exception"):
                exc_message = exc_info.get("message")
                exc_type = exc_info.get("type")
                exc_stacktrace = exc_info.get("stacktrace")
                message += f"\n{exc_type}: {exc_message}\n{exc_stacktrace}"

            code_info = nested_attrs.get("code", {})
            filename = code_info.get("file", {}).get("path")
            lineno = code_info.get("line", {}).get("number")
            function_name = code_info.get("function", {}).get("name")

            otel_info = nested_attrs.get("otel", {})
            logger_name = otel_info.get("library", {}).get("name")

            timestamp = datetime.fromisoformat(
                log_fields["timestamp"].replace("Z", "+00:00")
            )

            severity = log_fields.get("status", "info").upper()
            log_severity = (
                LoggingLevels[severity]
                if severity in LoggingLevels.__members__
                else LoggingLevels.INFO
            )

            module = None
            if function_name:
                module = function_name
            elif filename:
                module = filename.rsplit("/", 1)[-1].replace(".py", "")

            return LogEntry(
                message=message,
                level=log_severity,
                timestamp=timestamp,
                name=logger_name,
                filename=filename,
                lineno=lineno,
                module=module,
            )
        except Exception as e:
            logger.warning(f"Failed to parse log entry: {e}")
            return None

    def cleanup(self) -> None:
        """Cleanup the Datadog log store.

        This method is called when the log store is no longer needed.
        """
        if self._datadog_exporter:
            self._datadog_exporter.shutdown()
            self._datadog_exporter = None



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Base SecretSchema class."""

from typing import Any, Dict, List

from pydantic import BaseModel, ConfigDict


class BaseSecretSchema(BaseModel):
    """Base class for all Secret Schemas."""

    @classmethod
    def get_schema_keys(cls) -> List[str]:
        """Get all attributes that are part of the schema.

        These schema keys can be used to define all required key-value pairs of
        a secret schema.

        Returns:
            A list of all attribute names that are part of the schema.
        """
        return list(cls.model_fields.keys())

    def get_values(self) -> Dict[str, Any]:
        """Get all values of the secret schema.

        Returns:
            A dictionary of all attribute names and their corresponding values.
        """
        return self.model_dump(exclude_none=True)

    model_config = ConfigDict(
        # validate attribute assignments
        validate_assignment=True,
        # report extra attributes as validation failures
        extra="ignore",
    )



"""add visualisations [24552f3be1f2].

Revision ID: 24552f3be1f2
Revises: 279d55228d28
Create Date: 2025-10-24 07:56:57.575675

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "24552f3be1f2"
down_revision = "279d55228d28"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "curated_visualization",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("project_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column(
            "artifact_visualization_id",
            sqlmodel.sql.sqltypes.GUID(),
            nullable=False,
        ),
        sa.Column(
            "display_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
        sa.Column("display_order", sa.Integer(), nullable=True),
        sa.Column(
            "layout_size", sqlmodel.sql.sqltypes.AutoString(), nullable=False
        ),
        sa.Column("resource_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column(
            "resource_type", sqlmodel.sql.sqltypes.AutoString(), nullable=False
        ),
        sa.ForeignKeyConstraint(
            ["artifact_visualization_id"],
            ["artifact_visualization.id"],
            name="fk_curated_visualization_artifact_visualization_id",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_curated_visualization_project_id_project",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "artifact_visualization_id",
            "resource_id",
            "resource_type",
            name="unique_curated_visualization_resource_link",
        ),
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("curated_visualization")
    # ### end Alembic commands ###
